#!python

import shutil
import os 
import importlib
from pathlib import Path
import argparse,yaml
from argparse import RawTextHelpFormatter
import traceback,time,copy,yaml,sys,uuid
import sys
import subprocess
import re
sys.path.append(Path(__file__).resolve().parent.parent.__str__()) ## this line is for development
import dtiplayground
import dtiplayground.dmri.common
from dtiplayground.config import INFO as info

logger=dtiplayground.dmri.common.logger.write 
color= dtiplayground.dmri.common.Color

### unit functions

def resolve_softwarepaths(spathobj, globalvars):
    if 'dtiplayground-tools' in globalvars:
        p = Path(globalvars['dtiplayground-tools']['path'])
        if p.exists():
            os.environ['DTIPLAYGROUNDTOOLS'] = p.resolve().__str__()

    if 'fsl' in globalvars:
        p = Path(globalvars['fsl']['path'])
        if p.exists():
            os.environ['FSL'] = p.resolve().__str__()

    if  'DTIPLAYGROUNDTOOLS' not in os.environ:
        os.environ['DTIPLAYGROUNDTOOLS']=os.path.expandvars("$HOME/.niral/dtiplayground-tools")
        tooldir=os.environ['DTIPLAYGROUNDTOOLS']
        if Path(tooldir).exists():
            logger("DTI Playground tools directory found at {}".format(tooldir), color.OK)
        else:
            logger("DTI Playground tools directory is set to {}".format(tooldir), color.WARNING)
            logger("DTI Playground tools directory not found, please install toolkits or specify software paths manually", color.WARNING)
    else:
        tooldir=os.environ['DTIPLAYGROUNDTOOLS']
        if Path(os.environ['DTIPLAYGROUNDTOOLS']).exists():
            
            logger("DTI playground tools are found in {}".format(tooldir),color.OK)
        else:
            logger("DTI playground tools are not found in {}".format(tooldir),color.ERROR)
    if 'FSL' not in os.environ:
        logger("FSL directory is not set in environment variable FSL",color.WARNING)
        logger("Make sure FSL directory is set in the software paths",color.WARNING)
    else:
        if Path(os.environ['FSL']).exists():
            logger("FSL directory is found at {}".format(os.environ['FSL']),color.OK)
        else:
            logger("FSL directory is set in environment variable FSL, but doesn't exist",color.WARNING)
    
    sp = spathobj['softwares']
    for k,v in sp.items():
        resolved_path = os.path.expandvars(v['path'])
        sp[k]['path'] = resolved_path
    spathobj['softwares'] = sp
    return spathobj

def initialize_logger(args):
    ## default log setting
    dtiplayground.dmri.common.logger.setLogfile(args.log)
    dtiplayground.dmri.common.logger.setTimestamp(not args.no_log_timestamp)
    dtiplayground.dmri.common.logger.setVerbosity(not args.no_verbosity)

def check_initialized(args):
    home_dir=Path(args.config_dir)
    if home_dir.exists():
        config_file=home_dir.joinpath('config.yml')
        environment_file=home_dir.joinpath('environment.yml')
        if not config_file.exists():  return False
        if not environment_file.exists(): return False
        return True
    else: 
        return False

def load_configurations(config_dir:str):
    ## reparametrization
    home_dir=Path(config_dir)
    ## Function begins
    config_filename=home_dir.joinpath("config.yml")
    environment_filename=home_dir.joinpath("environment.yml")
    config=yaml.safe_load(open(config_filename,'r'))
    environment=yaml.safe_load(open(environment_filename,'r'))
    return config,environment

### decorators
def after_initialized(func): #decorator for other functions other than command_init
    def wrapper(args):
        home_dir=Path(args.config_dir)
        if home_dir.exists() and home_dir.joinpath('config.yml').exists() and home_dir.joinpath('environment.yml').exists():
            config_file=home_dir.joinpath('config.yml')
            environment_file=home_dir.joinpath('environment.yml')
            initialize_logger(args)
            return func(args)
        else: 
            # raise Exception("Not initialized, please run init command at the first use")
            logger("Configuration directory is not found or some files are missing, commencing initialization ...",dtiplayground.dmri.common.Color.WARNING)
            command_init(args)
            config_file=home_dir.joinpath('config.yml')
            environment_file=home_dir.joinpath('environment.yml')
            initialize_logger(args)
            return func(args)
    return wrapper

def log_off(func):
    def wrapper(*args,**kwargs):
        dtiplayground.dmri.common.logger.setVerbosity(False)
        res=func(*args,**kwargs)
        dtiplayground.dmri.common.logger.setVerbosity(True)
        return res 
    return wrapper

def search_fsl(lookupdirs=None):
    fslpath=None
    if 'FSLDIR' in os.environ:
        fslpath1=Path(os.environ['FSLDIR'])
        if fslpath1.exists():
            fslpath=fslpath1
    elif 'FSL' in os.environ:
        fslpath1=Path(os.environ['FSL'])
        if fslpath1.exists():
            fslpath=fslpath1
    else:
        fslpath=None

    if fslpath is not None:
        if fslpath.exists():
            logger("FSL directory found : {}".format(fslpath.resolve().__str__()), color.OK)
    else:
        logger("FSL directory not found for environment variable FSLDIR/FSL", color.WARNING)
        search_dirs=[]
        if lookupdirs is None or len(lookupdirs) < 1:
            search_dir = input("Specify the directory of installed FSL [/] :")
            if not search_dir: 
                search_dirs = ['/']
            else:
                search_dirs = [search_dir]
        else:
            logger("Looking up in the directories : {}".format(lookupdirs))
            search_dirs=lookupdirs

        for sdir in search_dirs:
            should_break=False
            search_dir = Path(sdir)
            candidates = search_dir.glob("**/bin/fsl")
            logger("Searching FSL in {}".format(search_dir.__str__()),color.PROCESS)
            for c in candidates:
                
                c = Path(c)
                if c.__str__().lower().endswith('/bin/fsl'):
                    fslpath = c.resolve().parent.parent
                    version=open(fslpath.joinpath('etc/fslversion'),'r').read().strip()
                    logger("Found : {}, Version: {}".format(fslpath.__str__(), version), color.OK)
                    confirm = input("Is this FSL you want? [Y/n]")
                    if not confirm: confirm = 'y'
                    if confirm.lower() == 'y':
                        should_break=True
                        break
            if should_break:
                break

    if fslpath.joinpath('etc/fslversion').exists():
        info = {
            "name" : "fsl",
            "version" : open(fslpath.joinpath('etc/fslversion'),'r').read().strip()
        }
        res = {
            "fsl" : {
                "path" : fslpath.resolve().__str__(),
                "info" : info
            }
        }
        return res
    else:
        return None

def search_dtiplayground_tools(lookupdirs=None):   
    target_path=None
    search_dirs=[]
    if lookupdirs is None or len(lookupdirs) < 1:
        search_dir = input("Specify the directory of installed DTI Playground tools [/] :")
        if not search_dir: 
            search_dirs = ['/']
        else:
            search_dirs = [search_dir]
    else:
        search_dirs=lookupdirs
    for sdir in search_dirs:
        should_break=False
        search_dir = Path(sdir)
        candidates = search_dir.glob("**/info.yml")
        info={}
        logger("Searching DTIPlaygroundTools in {}".format(search_dir.__str__()),color.PROCESS)
        for c in candidates:
            c = Path(c)
            info = yaml.safe_load(open(c,'r'))
            if 'name' in info:
                if info['name']=='dtiplayground-tools':
                    target_path = c.resolve().parent
                    version=info['version']
                    logger("Found : {}, Version: {}".format(target_path.__str__(), version), color.OK)
                    confirm = input("Is this DTI Playground Tools you want? [Y/n]")
                    if not confirm: confirm = 'y'
                    if confirm.lower() == 'y':
                        should_break=True
                        break
        if should_break: break

    if target_path is not None:
        res = {
            "dtiplayground-tools" : {
                "path" : target_path.resolve().__str__(),
                "info" : info
            }
        }
        return res
    else:
        return None
    
def update_software_paths(args, globalvars):
    home_dir=Path(args.config_dir)
    software_info_path=Path(dtiplayground.__file__).resolve().parent.joinpath("dmri/common/data/software_paths.yml")
    software_paths=yaml.safe_load(open(software_info_path,'r'))
    software_paths=resolve_softwarepaths(software_paths, globalvars)
    software_filename=home_dir.joinpath('software_paths.yml')
    yaml.dump(software_paths, open(software_filename,'w'))

### command functions

def command_find_tools(args):
    ## reparametrization
    home_dir=Path(args.config_dir)

    ## Function begins
    lookup_dirs = args.lookup_dirs
    config_filename=home_dir.joinpath("config.yml")
    environment_filename=home_dir.joinpath("environment.yml")
    software_filename=home_dir.joinpath('software_paths.yml')
    software_info_path=Path(dtiplayground.__file__).resolve().parent.joinpath("dmri/common/data/software_paths.yml")
    globalvars_fn = home_dir.parent.joinpath('global_variables.yml')
    globalvars={}
    if globalvars_fn.exists():
        globalvars.update(yaml.safe_load(open(globalvars_fn,'r')))

    ## load default software paths

    software_paths=yaml.safe_load(open(software_info_path,'r'))
    result = False
    ## search FSL
    fsl_info = search_fsl(lookupdirs=lookup_dirs)
    if fsl_info is not None:
        globalvars.update(fsl_info) 
        logger("Selected FSL directory is : {}".format(fsl_info['fsl']['path']),color.INFO)
    else:
        logger("FSL was not found",color.ERROR)
    ## search DTI Playground Tools
    
    dpt_info = search_dtiplayground_tools(lookupdirs=lookup_dirs)
    if dpt_info is not None:
        globalvars.update(dpt_info)
        logger("Selected DTI Playground tools directory is : {}".format(dpt_info['dtiplayground-tools']['path']),color.INFO)
    else:
        logger("DTI Playground tools were not found",color.ERROR)
    ## update software paths
    logger("Updating global_variables.yml", color.PROCESS)
    yaml.dump(globalvars, open(globalvars_fn,'w'))
    logger("Updating software paths",color.PROCESS)
    update_software_paths(args, globalvars)
    logger("Updates Completed",color.OK)
    result= True
    return result

def command_install_tools(args):
    ## reparametrization
    home_dir=Path(args.config_dir).resolve()
    home_dir.parent.mkdir(parents=True,exist_ok=True)
    # globalvars_fn = Path(os.path.expandvars("$HOME/.niral-dti/global_variables.yml"))
    globalvars_fn = home_dir.parent.joinpath('global_variables.yml')
    Path(globalvars_fn).parent.mkdir(parents=True,exist_ok=True)
    globalvars = {}
    if globalvars_fn.exists():
        globalvars = yaml.safe_load(open(globalvars_fn,'r'))

    rootdir = Path(os.path.expandvars(args.output_dir)).resolve()
    rootdir.mkdir(parents=True,exist_ok=True)
    tempdir = rootdir.joinpath('temp-dtiplayground-tools')
    outputdir = rootdir.joinpath('dtiplayground-tools')
    fsldir = rootdir.joinpath('FSL')
    clean_install = args.clean_install
    leave_files_after_installation = args.no_remove
    no_fsl = args.no_fsl
    install_only = args.install_only
    ### Clean install preparations
    if clean_install:
        logger("Removing temporary, existing packages (dtiplayground-tools only) directories ...")
        if tempdir.exists():
            shutil.rmtree(tempdir)
        if outputdir.exists():
            shutil.rmtree(outputdir)
        logger("Removing files finished")

    ### FSL
    if not no_fsl and not fsldir.exists():
        fslinstaller_fn=Path(__file__).resolve().parent.joinpath('fslinstaller.py')
        command=['/usr/bin/python', fslinstaller_fn ,'-d',fsldir.resolve().__str__()] ### python2 is needed
        subprocess.run(command)
        info = {
            'name' : 'fsl',
            'version' : open(fsldir.joinpath('etc/fslversion'),'r').read().strip()
        }
        globalvars['fsl']={
            'path' : fsldir.resolve().__str__(),
            'info' : info
        }
        yaml.dump(info, open(fsldir.joinpath('info.yml'),'w'))
    else:
        if fsldir.exists():
            info = {
                'name': 'fsl',
                'version' : open(fsldir.joinpath('etc/fslversion'),'r').read().strip()
            }
            yaml.dump(info, open(fsldir.joinpath('info.yml'),'w'))
            globalvars['fsl']={
                'path' : fsldir.resolve().__str__(),
                'info' : info
            }            

    ### DTI playground tools (build with docker, centos7 for now)
    tempdir.mkdir(parents=True,exist_ok=True)
    os.chdir(tempdir)
    srcdir = tempdir.joinpath('dtiplaygroundtools')
    info={}
    if not outputdir.exists():
        if not srcdir.exists():
            logger("Fetching source code ...")
            fetch_dtiplaygroundtools = ["git","clone","https://github.com/niraluser/dtiplaygroundtools.git"]
            subprocess.run(fetch_dtiplaygroundtools)
            logger("Source codes downloaded")
        
        os.chdir(srcdir.joinpath('dockerfiles'))
        build_command = ['./build.sh']
        logger("Building software packages ... ")
        subprocess.run(build_command)
        tar_filename = srcdir.joinpath('dist/dtiplayground-tools.tar.gz')
        untar_command = ['tar','xvfz',tar_filename.resolve().__str__(), '-C', rootdir.resolve().__str__()]
        logger("Installing softwares to the output directory {}".format(outputdir.resolve().__str__()))
        subprocess.run(untar_command)
        ### read version info
        info={}
        info_fn = outputdir.joinpath('info.yml')
        if info_fn.exists():
            info = yaml.safe_load(open(info_fn,'r'))

    else:
        info = yaml.safe_load(open(outputdir.joinpath('info.yml'),'r'))
    globalvars['dtiplayground-tools']={
        'path' : outputdir.resolve().__str__(),
        'info' : info
    }

    if not leave_files_after_installation:
        logger("Removing temporary files ...")
        shutil.rmtree(tempdir)
        logger("Removing temporary files finished")

    logger("Saving tool information to {}".format(globalvars_fn.__str__()))
    yaml.dump(globalvars, open(globalvars_fn,'w'))
    if home_dir.joinpath('software_paths').exists() and not install_only:
        update_software_paths(args, globalvars)
    logger("Installation completed")
    return True


def command_init(args):
    ## reparametrization
    home_dir=Path(args.config_dir)

    ## Function begins
    home_dir.mkdir(parents=True,exist_ok=True)
    tools_dir = None
    if args.tools_dir is not None:
        tools_dir = Path(args.tools_dir)
    user_module_dir=home_dir.parent.joinpath('modules/dmriprep').absolute()
    user_module_dir.mkdir(parents=True,exist_ok=True)
    user_tools_param_dir=home_dir.joinpath('parameters').absolute()
    user_tools_param_dir.mkdir(parents=True,exist_ok=True)
    template_filename=home_dir.joinpath("protocol_template.yml").absolute()
    source_template_path=Path(dtiplayground.dmri.preprocessing.__file__).parent.joinpath("templates/protocol_template.yml")
    protocol_template=yaml.safe_load(open(source_template_path,'r'))
    globalvars_fn=home_dir.parent.joinpath('global_variables.yml')
    globalvars = {}

    if globalvars_fn.exists():
        globalvars = yaml.safe_load(open(globalvars_fn,'r'))

    if tools_dir is not None:
        if tools_dir.exists():
            dptdir = tools_dir.joinpath('dtiplayground-tools')
            fsldir = tools_dir.joinpath('FSL')
            globalvars.update({
                "dtiplayground-tools" :{
                    "path" : dptdir.resolve().__str__(),
                    "info" : yaml.safe_load(open(dptdir.joinpath('info.yml'),'r'))
                },
                "fsl" : {
                    "path" : fsldir.resolve().__str__(),
                    "info" : {
                        'name' : 'fsl',
                        "verson" : open(fsldir.joinpath('etc/fslversion'),'r').read().strip()
                    }
                }
            })
            yaml.dump(globalvars, open(globalvars_fn,'w'))
        else:
            logger("Couldn't find tool directory {}".format(str(tools_dir)),color.ERROR)
            exit(1)

    initialize_logger(args)
    config_filename=home_dir.joinpath("config.yml")
    environment_filename=home_dir.joinpath("environment.yml")
    ## make configuration file (config.yml)
    config={"user_module_directories": [str(user_module_dir)],"protocol_template_path" : 'protocol_template.yml'}
    yaml.dump(protocol_template,open(template_filename,'w'))
    yaml.dump(config,open(config_filename,'w'))
    logger("Config file written to : {}".format(str(config_filename)),color.INFO)
    ## copy default software path
    software_filename=home_dir.joinpath('software_paths.yml')
    update_software_paths(args, globalvars)
    logger("Software path file is written to : {}".format(str(software_filename)),color.INFO)

    modules=dtiplayground.dmri.preprocessing.modules.load_modules(user_module_paths=config['user_module_directories'])
    environment=dtiplayground.dmri.preprocessing.modules.generate_module_envionrment(modules,str(home_dir))
    yaml.dump(environment,open(environment_filename,'w'))
    logger("Environment file written to : {}".format(str(environment_filename)),color.INFO)
    logger("Initialized. Local configuration will be stored in {}".format(str(home_dir)),color.OK)
    
    return True


@after_initialized
def command_update(args):
    ## reparametrization
    home_dir=Path(args.config_dir)

    ## Function begins
    config_filename=home_dir.joinpath("config.yml")
    environment_filename=home_dir.joinpath("environment.yml")

    ## load config_file
    config=yaml.safe_load(open(config_filename,'r'))
    ## make environment file (environment.yml)
    modules=dtiplayground.dmri.preprocessing.modules.load_modules(user_module_paths=config['user_module_directories'])
    environment=dtiplayground.dmri.preprocessing.modules.generate_module_envionrment(modules,str(home_dir))
    yaml.dump(environment,open(environment_filename,'w'))
    logger("Environment file written to : {}".format(str(environment_filename)),color.INFO)
    logger("Initialized. Local configuration will be stored in {}".format(str(home_dir)),color.OK)
    return True

def check_if_module_exists(args,module_name):
    home_dir=Path(args.config_dir)
    config_filename=home_dir.joinpath("config.yml")
    environment_filename=home_dir.joinpath("environment.yml")
    config=yaml.safe_load(open(config_filename,'r'))
    module_name = args.name
    system_module_root_dir = Path(dtiplayground.__file__).resolve().parent.joinpath("dmri/preprocessing/modules")
    user_module_root_dir = Path(config['user_module_directories'][0])
    if system_module_root_dir.joinpath(module_name).exists():
        return True
    if user_module_root_dir.joinpath(module_name).exists():
        return True
    return False 

@after_initialized
def command_add_module(args):
    home_dir=Path(args.config_dir)
    config_filename=home_dir.joinpath("config.yml")
    environment_filename=home_dir.joinpath("environment.yml")
    config=yaml.safe_load(open(config_filename,'r'))
    module_name = args.name
    system_module_root_dir = Path(dtiplayground.__file__).resolve().parent.joinpath("dmri/preprocessing/modules")
    user_module_root_dir = Path(config['user_module_directories'][0])
    module_dir = user_module_root_dir.joinpath(module_name)
    base_module_name = args.base_module

    if check_if_module_exists(args, module_name):
        logger("Module {} exists in {} or {}".format(module_name, user_module_root_dir.__str__(), system_module_root_dir.__str__()), color.ERROR)
        return False

    if base_module_name is None: ### NEW module 
        logger("Making directories and files ... ", color.PROCESS)
        module_dir.mkdir(parents=True,exist_ok=False)
        module_template_fn =Path(dtiplayground.__file__).resolve().parent.joinpath("dmri/preprocessing/templates/module_template.py")
        module_template = open(module_template_fn,'r').read()
        module_data_fn = module_template_fn.parent.joinpath('module_template.yml')
        module_data = open(module_data_fn,'r').read()

        ### replacing variable
        regex=r"@MODULENAME@"
        module_py = re.sub(regex,module_name,module_template,0)
        module_yml = re.sub(regex,module_name,module_data,0)
        ### write down files into user module dir

        init_fn = module_dir.joinpath("__init__.py")
        with open(init_fn,'w') as f:
            f.write("")
        out_py_fn = module_dir.joinpath("{}.py".format(module_name))
        with open(out_py_fn,'w') as f:
            f.write(module_py)
        out_yml_fn = module_dir.joinpath("{}.yml".format(module_name))
        with open(out_yml_fn,'w') as f:
            f.write(module_yml)
        logger("Module is added, please implement the module {} in {}".format(module_name, module_dir.__str__()), color.OK)
    else: ### copy base module to new one (new name)
        base_module_dir= None
        if system_module_root_dir.joinpath(base_module_name).exists():
            base_module_dir = system_module_root_dir.joinpath(base_module_name)
            logger("Module {} found in system module directory".format(base_module_name),color.OK)
        elif user_module_root_dir.joinpath(base_module_name).exists():
            base_module_dir = user_module_root_dir.joinpath(base_module_name)
            logger("Module {} found in user module directory".format(base_module_name),color.OK)
        else:
            logger("Module {} NOT found in either system or user module directory".format(base_module_name),color.ERROR)
            return False

        logger("Generating {} at {}".format(module_name, module_dir.__str__()),color.PROCESS)
        shutil.copytree(base_module_dir, module_dir)
        if module_dir.joinpath('__pycache__').exists():
            shutil.rmtree(module_dir.joinpath('__pycache__'))
        flist = list(filter(lambda x: '__pycache__' not in x.__str__(), module_dir.glob("**/*")))
        for fn in flist: ### changing module name in contents (may not perfect)
            out_content=None
            with open(fn,'r') as f:
                ### change class name to new module name
                content = f.read()
                out_content = content.replace(base_module_name, module_name)
                out_content = out_content.replace(base_module_name.replace("_"," "), module_name.replace("_"," "))
            with open(fn,'w') as f:
                f.write(out_content)

            nameparts = fn.name.split('.')
            if nameparts[0] == base_module_name:
                nameparts[0] = module_name
            out_fn = ".".join(nameparts)
            fn.rename(module_dir.joinpath(out_fn))
        logger("Module {} cloned from {} has been generated at {}".format(module_name, base_module_name, module_dir.__str__()),color.OK)
        if args.edit:
            command=["vi",module_dir.joinpath("{}.py".format(module_name))]
            subprocess.run(command)
    return True

@after_initialized
def command_remove_module(args):
    home_dir=Path(args.config_dir)
    config_filename=home_dir.joinpath("config.yml")
    environment_filename=home_dir.joinpath("environment.yml")
    config=yaml.safe_load(open(config_filename,'r'))
    module_name = args.name
    module_root_dir = Path(config['user_module_directories'][0])
    module_dir = module_root_dir.joinpath(module_name)
    if module_dir.exists():
        logger("Removing the module : {} at {}".format(module_name, module_dir.__str__()), color.PROCESS)
        shutil.rmtree(module_dir)
        logger("Module {} has been successfully removed".format(module_name), color.OK)
    else:
        logger("There is no such module in {}".format(module_root_dir.__str__()), color.ERROR)

    return True

def parse_global_variables(global_vars: list):
    gv = {}
    n_vars = int(len(global_vars)/2)
    for i in range(n_vars):
        gv[global_vars[i]]=global_vars[i+1]
    return gv

@after_initialized
@log_off
def command_make_protocols(args):
    ## reparametrization
    options={
        "config_dir" : args.config_dir,
        "input_image_paths" : args.input_images,
        "module_list": args.module_list,
        "output_path" : args.output,
        "baseline_threshold" : args.b0_threshold,
        "output_format" : args.output_format,
        "no_output_image" : args.no_output_image,
        "global_variables" : parse_global_variables(args.global_variables)
    }
    if options['output_path'] is not None:
        dtiplayground.dmri.common.logger.setVerbosity(True)
    ## load config file
    config,environment = load_configurations(options['config_dir'])
    template_path=Path(options['config_dir']).joinpath(config['protocol_template_path'])
    template=yaml.safe_load(open(template_path,'r'))
    proto=dtiplayground.dmri.preprocessing.protocols.Protocols(options['config_dir'],global_vars=options['global_variables'])
    proto.loadImages(options['input_image_paths'],b0_threshold=options['baseline_threshold'])
    if options['module_list'] is not None and  len(options['module_list'])==0:
            options['module_list']=None
    proto.makeDefaultProtocols(options['module_list'],template=template,options=options)
    outstr=yaml.dump(proto.getProtocols())
    print(outstr)
    if options['output_path'] is not None:
        open(options['output_path'],'w').write(outstr)
        logger("Protocol file has been writte to : {}".format(options['output_path']),color.OK)


@after_initialized
def command_run(args):
    ## reparametrization
    options={
        "config_dir" : args.config_dir,
        "input_image_paths" : args.input_image_list,
        "protocol_path" : args.protocols,
        "output_dir" : args.output_dir,
        "default_protocols":args.default_protocols,
        "num_threads":args.num_threads,
        "execution_id":args.execution_id,
        "baseline_threshold" : args.b0_threshold,
        "output_format" : args.output_format,
        "output_file_base" : args.output_file_base,
        "no_output_image" : args.no_output_image,
        "global_variables" : parse_global_variables(args.global_variables)
    }
    if args.output_format is not None:
        options['output_format']=args.output_format.lower()
    ## load config file and run pipeline
    config,environment = load_configurations(options['config_dir'])
    template_path=Path(options['config_dir']).joinpath(config['protocol_template_path'])
    template=yaml.safe_load(open(template_path,'r'))
    proto=dtiplayground.dmri.preprocessing.protocols.Protocols(options['config_dir'], global_vars=options['global_variables'])
    proto.loadImages(options['input_image_paths'],b0_threshold=options['baseline_threshold'])
    if options['output_dir'] is None:
        raise Exception("Output directory is missing")
    else:
        proto.setOutputDirectory(options['output_dir'])
    if options['default_protocols'] is not None:
        if len(options['default_protocols'])==0:
            options['default_protocols']=None
        proto.makeDefaultProtocols(options['default_protocols'],template=template,options=options)
    elif options['protocol_path'] is not None:
        proto.loadProtocols(options["protocol_path"])
    else :
        proto.makeDefaultProtocols(options['default_protocols'],template=template,options=options)
    if options['num_threads'] is not None:
        proto.setNumThreads(options['num_threads'])
    Path(options['output_dir']).mkdir(parents=True,exist_ok=True)
    logfilename=str(Path(options['output_dir']).joinpath('log.txt').absolute())
    dtiplayground.dmri.common.logger.setLogfile(logfilename)  
    logger("\r----------------------------------- QC Begins ----------------------------------------\n")
    res=proto.runPipeline(options=options)
    logger("\r----------------------------------- QC Done ----------------------------------------\n")
    return res 
### Arguments 

def get_args():
    current_dir=Path(__file__).resolve().parent
    # info=yaml.safe_load(open(current_dir.parent.joinpath('dtiplayground/info.json'),'r'))
    version=info['dmriprep']['version']
    logger("VERSION : {}".format(str(version)))
    config_dir=Path(os.environ.get('HOME')).joinpath('.niral-dti/dmriprep-'+str(version))
    # ## read template
    module_help_str=None
    if config_dir.exists() and config_dir.joinpath('config.yml').exists() and config_dir.joinpath('environment.yml').exists():
        config,environment = load_configurations(str(config_dir))
        template_path=config_dir.joinpath(config['protocol_template_path'])
        template=yaml.safe_load(open(template_path,'r'))
        available_modules=template['options']['execution']['pipeline']['candidates']
        available_modules_list=["{}".format(x['value'])  for x in available_modules if x['description']!="Not implemented"]
        module_help_str="Avaliable Modules := \n" + " , ".join(available_modules_list)
        # print(module_help_str)
    uid, ts = dtiplayground.dmri.common.get_uuid(), dtiplayground.dmri.common.get_timestamp()

    ### Argument parsers

    parser=argparse.ArgumentParser(prog="dmriprep",
                                   formatter_class=RawTextHelpFormatter,
                                   description="dmriprep is a tool that performs quality control over diffusion weighted images. Quality control is very essential preprocess in DTI research, in which the bad gradients with artifacts are to be excluded or corrected by using various computational methods. The software and library provides a module based package with which users can make his own QC pipeline as well as new pipeline modules.",
                                   epilog="Written by SK Park (sangkyoon_park@med.unc.edu) , Johanna Dubos (johannadubos32@gmail.com) , Neuro Image Research and Analysis Laboratories, University of North Carolina @ Chapel Hill , United States, 2021")
    subparsers=parser.add_subparsers(help="Commands")
    
    ## init command
    parser_init=subparsers.add_parser('init',help='Initialize configurations')
    parser_init.add_argument('--tools-dir', help="Initialize with specific tool directory", default=None)
    parser_init.set_defaults(func=command_init)

    ## update command
    parser_update=subparsers.add_parser('update',help='Update environment file')
    parser_update.set_defaults(func=command_update)

    ## add new module
    parser_new_module=subparsers.add_parser('add-module', help='Add new module to user module directory')
    parser_new_module.add_argument('name', help="Module name")
    parser_new_module.add_argument('-b','--base-module', help="Fork from an existing module with new name", default=None, required=False)
    parser_new_module.add_argument('-e','--edit', help="Run vi editor after generating module", default=False, action="store_true")
    parser_new_module.set_defaults(func=command_add_module)


    ## remove user module 
    parser_remove_module=subparsers.add_parser('remove-module', help='Remove a user module from user module directory')
    parser_remove_module.add_argument('name', help="Module name")
    parser_remove_module.set_defaults(func=command_remove_module)

    ## software find command
    parser_find_tools=subparsers.add_parser('find-tools',help='Search and update software paths')
    parser_find_tools.add_argument('-l','--lookup-dirs', help='Lookup dir list', type=str, nargs='*',required=False)
    parser_find_tools.set_defaults(func=command_find_tools)

    ## software install command
    parser_install_tools=subparsers.add_parser('install-tools',help='Install DTIPlaygroundTools')
    parser_install_tools.add_argument('-o','--output-dir', help="output directory", default="$HOME/.niral-dti")
    parser_install_tools.add_argument('-c','--clean-install', help="Remove existing files", default=False, action="store_true")
    parser_install_tools.add_argument('--no-remove', help="Do not remove source and build files after installation", default=False,action="store_true")
    parser_install_tools.add_argument('--no-fsl', help="Do not install FSL", default=False, action="store_true")
    parser_install_tools.add_argument('--install-only', help="Do not update current software paths", default=False, action="store_true")
    parser_install_tools.set_defaults(func=command_install_tools)

    ## generate-default-protocols
    parser_make_protocols=subparsers.add_parser('make-protocols',help='Generate default protocols',epilog=module_help_str)
    parser_make_protocols.add_argument('-i','--input-images',help='Input image paths',type=str,nargs='+',required=True)
    parser_make_protocols.add_argument('-g','--global-variables',help='Global Variables',type=str,nargs='*',required=False)
    parser_make_protocols.add_argument('-o','--output',help='Output protocol file(*.yml)',type=str)
    parser_make_protocols.add_argument('-d','--module-list',metavar="MODULE",
                                        help='Default protocols with specified list of modules, only works with default protocols. Example : -d DIFFUSION_Check SLICE_Check',
                                        default=None,nargs='*')
    parser_make_protocols.add_argument('-b','--b0-threshold',metavar='BASELINE_THRESHOLD',help='b0 threshold value, default=10',default=10,type=float)
    parser_make_protocols.add_argument('-f','--output-format',metavar='OUTPUT FORMAT',default=None,help='OUTPUT format, if not specified, same format will be used for output (NRRD | NIFTI)',type=str)
    parser_make_protocols.add_argument('--no-output-image',help="No output Qced file will be generated",default=False,action='store_true')
    parser_make_protocols.set_defaults(func=command_make_protocols)
        

    ## run command
    parser_run=subparsers.add_parser('run',help='Run pipeline',epilog=module_help_str)
    parser_run.add_argument('-i','--input-image-list',help='Input image paths',type=str,nargs='+',required=True)
    parser_run.add_argument('-g','--global-variables',help='Global Variables',type=str,nargs='*',required=False)
    parser_run.add_argument('-o','--output-dir',help="Output directory",type=str,required=True)
    parser_run.add_argument('--output-file-base', help="Output filename base", type=str, required=False)
    parser_run.add_argument('-t','--num-threads',help="Number of threads to use",default=None,type=int,required=False)
    parser_run.add_argument('--no-output-image',help="No output Qced file will be generated",default=False,action='store_true')
    parser_run.add_argument('-b','--b0-threshold',metavar='BASELINE_THRESHOLD',help='b0 threshold value, default=10',default=10,type=float)
    parser_run.add_argument('-f','--output-format',metavar='OUTPUT FORMAT',default=None,help='OUTPUT format, if not specified, same format will be used for output  (NRRD | NIFTI)',type=str)
    run_exclusive_group=parser_run.add_mutually_exclusive_group()
    run_exclusive_group.add_argument('-p','--protocols',metavar="PROTOCOLS_FILE" ,help='Protocol file path', type=str)
    run_exclusive_group.add_argument('-d','--default-protocols',metavar="MODULE",help='Use default protocols (optional : sequence of modules, Example : -d DIFFUSION_Check SLICE_Check)',default=None,nargs='*')
    parser_run.set_defaults(func=command_run)

    ## log related
    parser.add_argument('--config-dir',help='Configuration directory',default=str(config_dir))
    parser.add_argument('--log',help='log file',default=str(config_dir.joinpath('log.txt')))
    # parser.add_argument('--system-log-dir',help='System log directory',default='/BAND/USERS/skp78-dti/system-logs',type=str)
    parser.add_argument('--execution-id',help='execution id',default=uid,type=str)
    parser.add_argument('--no-log-timestamp',help='Remove timestamp in the log', default=False, action="store_true")
    parser.add_argument('--no-verbosity',help='Do not show any logs in the terminal', default=False, action="store_true")
    parser.add_argument('-v','--version', help="Show version", default=False,action="store_true")

    ## if no parameter is furnished, exit with printing help
    if len(sys.argv)==1:
        parser.print_help(sys.stderr)
        sys.exit(1)
    args=parser.parse_args()
    if args.version:
        sys.exit(1)

    return args 

## threading environment
args=get_args()
if hasattr(args,'num_threads'):
    os.environ['OMP_NUM_THREADS']=str(args.num_threads) ## this should go before loading any dipy function. 
    os.environ['ITK_GLOBAL_DEFAULT_NUMBER_OF_THREADS'] = str(args.num_threads) ## for ANTS threading

import dtiplayground.dmri.preprocessing
import dtiplayground.dmri.preprocessing.modules
import dtiplayground.dmri.preprocessing.protocols

if __name__=='__main__':
    try:
        dtiplayground.dmri.common.logger.setTimestamp(True)
        result=args.func(args)
        exit(0)
    except Exception as e:
        dtiplayground.dmri.common.logger.setVerbosity(True)
        msg=traceback.format_exc()
        logger(msg,color.ERROR)
        exit(-1)
    finally:
        pass







from .brainsfit import BRAINSFit 
from .dti_reg import DTIReg 
from .dtiaverage import DTIAverage
from .dtiprocess import DTIProcess
from .greedy_atlas import GreedyAtlas
from .image_math import ImageMath
from .itk_transform_tools import ITKTransformTools 
from .resample_dti_log_euclidean import ResampleDTIlogEuclidean
from .unu import UNU 
from .crop_dti import CropDTI
from .fsl import FSL
from .dtiestim import DTIEstim






  
import dtiplayground.dmri.preprocessing as prep
from dtiplayground.dmri.common import measure_time
import dtiplayground.dmri.common.tools as tools 
from dtiplayground.dmri.preprocessing.dwi import DWI
import yaml
from pathlib import Path
import copy

import numpy as np
import dipy.reconst.dti as dti
from dipy.core.gradients import gradient_table
import dipy.denoise.noise_estimate as ne
from dipy.io.image import save_nifti

logger=prep.logger.write


class DTI_Estimate(prep.modules.DTIPrepModule):
    def __init__(self,config_dir,*args,**kwargs):
        super().__init__(config_dir,*args,**kwargs)
        
    def generateDefaultProtocol(self,image_obj):
        super().generateDefaultProtocol(image_obj)
        ## todos
        return self.protocol

    def process(self,*args,**kwargs): ## variables : self.global_variables, self.softwares, self.output_dir, self.config_dir, self.source_image, self.image (output) , self.result_history , self.result (output) , self.protocol, self.template
        super().process()
        inputParams=self.getPreviousResult()['output']
        # << TODOS>>
        protocol_options=args[0]
        self.num_threads=protocol_options['software_info']['parameters']['num_max_threads']
        self.software_info=protocol_options['software_info']['softwares']
        self.baseline_threshold=protocol_options['baseline_threshold']
        self.global_vars=self.getGlobalVariables()
        res=self.runDTI(method=self.protocol['method'],
                        optimizationMethod=self.protocol['optimizationMethod'],
                        correctionMethod=self.protocol['correctionMethod'])
        self.result['output']['success']=True
        return self.result

### User defined methods

    def runDTI(self,method, optimizationMethod, correctionMethod):
        res = None
        if method.lower() == 'dipy':
            logger("Using {}".format(method),prep.Color.INFO)
            res = self.runDTI_DIPY(optimizationMethod)
        elif method.lower() == 'dtiestim':
            logger("Using {}".format(method),prep.Color.INFO)
            res = self.runDTI_dtiestim(optimizationMethod,correctionMethod)
        else:
            raise Exception("Unknown method name : {}".format(method))
    
    @measure_time
    def runDTI_DIPY(self, optimizationMethod):
        
        # data prep for dipy
        data = self.image.images
        affine = self.image.getAffineMatrixForNifti()
        bvals= np.array(list(map(lambda x: x['b_value'], self.image.getGradients())))
        b0=min(bvals)
        bvecs= np.array(list(map(lambda x: x['unit_gradient'], self.image.getGradients())))
        gtab = gradient_table(bvals,bvecs,b0_threshold=min(max(b0,50),199))
        logger("Affine Matrix (RAS) : \n{}".format(affine),prep.Color.INFO)
        # option parse
        optionmap = { 'lls' : 'LS',
                      'wls' : 'WLS',
                      'nls': 'NLLS',
                      'restore' : 'RESTORE' }
        fitMethod="WLS"
        kwargs={}
        try:
            fitMethod=optionmap[optimizationMethod]
            if fitMethod=='RESTORE':
                kwargs={ 'sigma': ne.estimate_sigma(data) }
        except:
            fitMethod="WLS"
            logger("WARNING: The method {} is not available with the method. Changing it to {}.".format(optimizationMethod,fitMethod), prep.Color.WARNING)

        # Try loading mask if exists
        mask=None
        if 'mask_path' in self.global_vars:
            mpath = self.global_vars['mask_path']
            if Path(mpath).exists():
                logger('Mask file found : {}'.format(self.global_vars['mask_path']),prep.Color.OK)
                temp_img=DWI(mpath)
                mask=temp_img.images
                # print(mask)
                # print(mask.shape)
            else:
                logger('Mask not found, estimating whole image...',prep.Color.WARNING)
        else:
            logger('Mask not found, estimating whole image...',prep.Color.WARNING)
        # fitting and estimation of scalars
        dti_fit = dti.TensorModel(gtab,fit_method=fitMethod,**kwargs)
        logger("Running with {}, {}".format(fitMethod, kwargs),prep.Color.PROCESS)
        #fitted = dti_fit.fit(data) ## dti_fit.fit(data, mask) mask array (boolean)
        fitted = dti_fit.fit(data,mask)
        logger("Fitting completed",prep.Color.OK)

        ## convert 3x3 symmetric matrices to xx,xy,xz,yy,yz,zz vectors
        logger("Reducing 3x3 symmetric matrix to vector")
        def uppertriangle(matrix):
            outvec=[]
            for i in range(3):
                for j in range(i,3):
                    outvec.append(matrix[i,j])
            return np.array(outvec)
        quad_form = fitted.quadratic_form
        new_quadform = np.ndarray(shape=(quad_form.shape[0],quad_form.shape[1],quad_form.shape[2],6),dtype=float)
        for d1 in range(quad_form.shape[0]):
            for d2 in range(quad_form.shape[1]):
                for d3 in range(quad_form.shape[2]):
                    mat = quad_form[d1,d2,d3]
                    new_quadform[d1,d2,d3]=uppertriangle(mat)

        # TODO : make nrrd file for new_quadform image volume (kind will be "3D-symmetric-matrix") , ref: http://teem.sourceforge.net/nrrd/format.html
        temp_dti_image = copy.deepcopy(self.image)
        temp_dti_image.setImage(new_quadform,modality='DTI', kinds=['space','space','space','3D-symmetric-matrix'])
        dti_filename=Path(self.output_dir).joinpath('tensor.nrrd').__str__()
        sp_dir=self.getSourceImageInformation()['space']
        temp_dti_image.setSpaceDirection(target_space=sp_dir)
        temp_dti_image.writeImage(dti_filename,dest_type='nrrd',dtype="float32")
        self.addOutputFile(dti_filename, 'DTI')
        self.addGlobalVariable('dti_image',dti_filename)
        # retrieve outputs
        evals = fitted.evals
        evecs = fitted.evecs
        scalarData={
            'eigenval': evals,
            'eigenvec': evecs,
            'fa': fitted.fa,
            'cfa': dti.color_fa(fitted.fa,evecs),
            'md': fitted.md,
            'ad': fitted.ad,
            'rd': fitted.rd    
        }
        # sphere = dpd.default_sphere

        # saving outputs
        for scalar, val in scalarData.items():
            output_tensor_path = Path(self.output_dir).joinpath('tensor_{}.nii.gz'.format(scalar)).__str__()
            val[np.isnan(val)] = 0
            num_type=np.float32
            save_nifti(output_tensor_path, val.astype(num_type), affine)
            self.addOutputFile(output_tensor_path, 'DTI_{}'.format(scalar.upper()))

        return None

    @measure_time
    def runDTI_dtiestim(self, optimizationMethod, correctionMethod):

        # option parse
        optionmap = { 'lls' : 'lls',
                      'wls' : 'wls',
                      'nls': 'nls',
                      'ml' : 'ml'}
        fitMethod="wls"
        try:
            fitMethod=optionmap[optimizationMethod]
        except:
            fitMethod="wls"
            logger("WARNING: The method {} is not available with the method. Changing it to {}.".format(optimizationMethod,fitMethod), prep.Color.WARNING)


        dtiestim=tools.DTIEstim(self.software_info['dtiestim']['path'])
        input_image_path = Path(self.output_dir).joinpath('input.nrrd').__str__()
        output_tensor_path = Path(self.output_dir).joinpath('tensor.nrrd').__str__()
        self.writeImage(str(input_image_path),dest_type='nrrd')

        options = [ '-m',fitMethod,
                    '--correction', correctionMethod]

        dtiestim.estimate(input_image_path, output_tensor_path,options)
        self.addOutputFile(output_tensor_path, 'DTI')
        self.addGlobalVariable('dti_image',output_tensor_path)
        return None




import dtiplayground.dmri.preprocessing as prep

import numpy as np
import nrrd
import nibabel as nib

import yaml
from pathlib import Path
import copy

#
#
# gradients are unit-vectors normalized according to b-value (nrrd), there is also normalized gradient coupled with bvalue
#
#

logger=prep.logger.write

def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]

def _load_nrrd(filename):
    org_data,header = nrrd.read(filename)
    
    ## dimension checkout
    kinds=[]
    grad_axis=0
    for idx,k in enumerate(header['kinds']):
        if k.lower() != "domain" and k.lower() != 'space':
            grad_axis=idx 
            kinds.append(False)
        else:
            kinds.append(True)
    ## extract image size 
    img_size=list(header['sizes'])
    space_directions=[]
    grad_size=None
    for idx,k in enumerate(kinds):
        if k: 
            # img_size.append(list(header['sizes'].tolist()))
            space_directions.append(header['space directions'].tolist()[idx])
        else:
            grad_size=header['sizes'][idx]

    info={
        'space':header['space'],
        'dimension': int(header['dimension']),
        'sizes': list(map(int, img_size)), #header['sizes'],
        'original_kinds': header['kinds'],
        'original_kinds_space' : kinds,
        'image_size' : list(map(int,img_size[:3])),
        # 'b_value':float(header['DWMRI_b-value']),
        'space_directions': space_directions,
        'space_origin':header['space origin'].tolist(),
        #'original_centerings': header['centerings'],
        'endian' : header['endian'],
        'type' : header['type']
    }
    if 'measurement_frame' in header:
        info['measurement_frame'] = header['measurement_frame'].tolist()
    else:
        info['measurement_frame'] =  np.identity(3).tolist()
    if 'modality' in header:
        info['modality'] = header['modality']
    else:
        info['modality'] = None

    if 'DWMRI_b-value' in header:
        info['b_value'] = float(header['DWMRI_b-value'])
    else:
        info['b_value'] = None
    if 'centerings' in header:
        info['original_centerings']=header['centerings']
    else:
        info['original_centerings']=None 

    if "thicknesses" in header :
        info['thicknesses'] = header['thicknesses'].tolist()
    else:
        info["thicknesses"] = None

    data=org_data.copy()
    ### move axis to match nifti
    if 'DWMRI_b-value' in header: ## only for the dwmri images
        data=np.moveaxis(data,grad_axis,-1)
        info['sizes']=list(data.shape)
        info['image_size']=list(data.shape[0:3])
    ### extracting gradients
    gradients=[]
    for k,v in header.items():
        if 'DWMRI_gradient' in k:
            idx=int(k.split('_')[2])
            vec=np.array(list(map(lambda x: float(x),v.split())))
            bval=float(np.sum(vec**2)*info['b_value'])
            normalize_term=np.sqrt(np.sum(vec**2))
            if normalize_term>0:
                unit_vec=(vec/normalize_term)
            else:
                unit_vec=vec 
            gradients.append({'index':idx,'gradient':vec.tolist(),'b_value':bval,'unit_gradient':unit_vec.tolist(),'original_index':idx})
    gradients=sorted(gradients,key=lambda x: x['index'])
    return data,gradients,info , (org_data,header)

def _load_nifti_bvecs(filename):
    content=open(filename,'r').read().strip()
    num_gradients=len(content.split())/3
    is_old_format= len(content.split('\n'))!=num_gradients 
    if is_old_format:
        x,y,z=content.strip().split('\n')
        x=x.split()
        y=y.split()
        z=z.split()
        return list(zip(x,y,z))
    else:
        return list(chunks(content.split(),3))

def _load_nifti(filename,bvecs_file=None,bvals_file=None):
    parent_dir=Path(filename).parent
    if bvals_file is None: bvals_file=parent_dir.joinpath(Path(Path(filename).stem).stem+'.bval')
    if bvecs_file is None: bvecs_file=parent_dir.joinpath(Path(Path(filename).stem).stem+'.bvec')

    if not bvals_file.exists():
        bvals_file=parent_dir.joinpath(Path(Path(filename).stem).stem+'.bvals')
    if not bvecs_file.exists():
        bvecs_file=parent_dir.joinpath(Path(Path(filename).stem).stem+'.bvecs')
    
    loaded_image_object= nib.load(filename)
    header=loaded_image_object.header
    org_data=loaded_image_object.get_fdata().astype(np.dtype(header.get_data_dtype()))
    image_dim = len(org_data.shape)
    gradients=None

    ## extract gradients with form {'index': , 'gradient': }
    bvals=[]
    bvecs=[]
    gradients=[]
    max_bval=0.0

    if image_dim == 4:
        tmp_bvals=list(open(bvals_file,'r').read().split())
        tmp_bvecs=_load_nifti_bvecs(bvecs_file)
        bvpairs=list(zip(tmp_bvals,tmp_bvecs))
        normalized_vecs=[]
        vecs=[]
        
        for idx,bv in enumerate(bvpairs):
            bval,bvec = bv
            normalized_vecs.append(list(map(lambda x :float(x),bvec)))
            bvals.append(float(bval))


        max_bval=np.max(bvals)
        for idx,vec in enumerate(normalized_vecs):
            denormalized_vec=np.array(vec)*np.sqrt((bvals[idx]/max_bval))
            gradients.append({'index':int(idx),
                              'gradient': denormalized_vec.tolist(),
                              'b_value': bvals[idx],
                              'unit_gradient': vec,
                              'original_index':idx})

    ## move gradient index to the first (same to nrrd format)
    

    affine=loaded_image_object.affine 
    ijk_to_lps = affine
    lps_to_ras = np.diag([-1, -1, 1, 1]) # ras to lps 
    ijk_to_ras = np.matmul(lps_to_ras, ijk_to_lps)
    affine=ijk_to_ras

    data=org_data.copy()

    ## extract header info
    mat=np.array(affine)

    space='left-posterior-superior'

    space_directions=mat.transpose()[:3,:3]
    space_origin=mat.transpose()[3,:3]

    endian="little"
    if header.endianness != '<' :
        endian='big'

    info={
        'space': space,
        'dimension': len(data.shape),
        'sizes': np.array(data.shape).tolist(),
        "original_kinds": ['space','space','space','list'][:image_dim],
        "original_kinds_space" : [True,True,True,False][:image_dim], ## image space = True, gradient dim = False
        'image_size' : np.array(data.shape[0:3]).tolist(),
        'b_value': float(max_bval),
        'space_directions': space_directions.tolist(),
        'measurement_frame': np.identity(3).tolist(), ## this needs to be clarified for nifti
        'space_origin':space_origin.tolist(),
        'type': str(header.get_data_dtype()),
        'endian' : endian,
        'original_centerings' : ['cell','cell','cell','???'][:image_dim],
        'thicknesses' : np.array([np.NAN,np.NAN,np.abs(space_directions.tolist()[2][2]),np.NAN]).tolist()[:image_dim],
        'modality': None
    }



    return data, gradients, info, (org_data,affine,header)

def _load_dwi(filename, filetype='nrrd'):
    if filetype.lower()=='nrrd': ## load nrrd dwi image
        return _load_nrrd(filename)
    elif filetype.lower()=='nifti':
        return _load_nifti(filename)
    else:
        logger("Not a supported image type",prep.Color.ERROR)
        return None

def get_nrrd_gradient_axis(kinds):
    grad_axis=0
    for idx,k in enumerate(kinds):
        if k.lower() != "domain" and k.lower() != 'space':
            grad_axis=idx 
    return grad_axis



def export_to_nrrd(image): #image : DWI
    info=copy.deepcopy(image.information)
    grad=image.getGradients()
    new_data=copy.deepcopy(image.images)
    grad_axis_original=get_nrrd_gradient_axis(info['original_kinds'])
    grad_axis=-1
    
    space_directions=info['space_directions']
    if image.information['dimension']>3:
        space_directions.append([np.NAN,np.NAN,np.NAN])
        space_directions_grad_axis=space_directions[grad_axis]
        space_directions=np.insert(space_directions,
                                    grad_axis_original,
                                    space_directions_grad_axis,
                                    axis=0)
        space_directions=np.delete(space_directions,-1,axis=0)
        space_directions=space_directions.tolist()
        

    new_header={
        "type": info['type'],
        "dimension": info['dimension'],
        "space":  info['space'],
        "sizes":  info['sizes'],
        "space directions": space_directions,
        "kinds": info['original_kinds'],
        "endian" : info['endian'],
        "encoding" : 'gzip',
        "space origin" : info['space_origin'],
        "measurement frame": info['measurement_frame']
    }
    if 'modality' in info:
        new_header['modality']=info['modality']
    else:
        new_header['modality']=None
    if info['dimension'] == 4 and info['modality'] != 'DTI' and info['b_value'] is not None:
        new_header['modality']="DWMRI"
        new_header['DWMRI_b-value']=info['b_value']

    if info['original_centerings'] is not None:
        new_header["centerings"]= info['original_centerings'] 
    if info['thicknesses'] is not None:
        new_header['thicknesses']= info['thicknesses']


    s=list(new_data.shape)
    new_header['dimension']=len(new_data.shape)
    if new_header['dimension'] > 3:
        g=s[grad_axis]
        s=s[:grad_axis]+s[grad_axis:-1]
        s=s[:grad_axis_original]+[g]+s[grad_axis_original:]
    new_header['sizes']=s

    copy_hdr=copy.deepcopy(new_header)
    for k,v in copy_hdr.items():    
        if 'dwmri_gradient' in k.lower():
            del new_header[k]
    if new_header['modality']=='DWMRI':
        for idx,g in enumerate(grad):
            k="DWMRI_gradient_{:04d}".format(idx)
            new_header[k]=" ".join([str(x) for x in g['gradient']])
        if new_header['dimension'] > 3:
            new_data=np.moveaxis(new_data,grad_axis,grad_axis_original)
    new_data=new_data.astype(new_header['type'])
    return new_data,new_header 


def export_to_nifti(image): # image DWI
    affine=image.getAffineMatrixForNifti()
    img=image.images 
    gradients=image.getGradients()
    bvals=["{:d}\n".format(int(x['b_value'])) for x in gradients]
    bvecs=[" ".join(map(lambda s : "{:.8f}".format(s),x['unit_gradient']))+"\n" for x in gradients]
    return img,affine, bvals, bvecs

def _write_nrrd(image,filename,dtype): 
    image.information['type']=dtype
    data,header = export_to_nrrd(image)
    return nrrd.write(filename,data,header=header)

def _write_nifti(image,filename,dtype): #image : DWI
    data,affine,bvals,bvecs=export_to_nifti(image)
    out_dir=Path(filename).parent
    filename_stem=Path(filename).name.split('.')[0]
    bvals_filename=out_dir.joinpath(filename_stem+".bval")
    bvecs_filename=out_dir.joinpath(filename_stem+".bvec")
    data=data.astype(dtype)
    out_image_object=nib.Nifti1Image(data,affine)
    nib.save(out_image_object,str(filename))
    ## wrting bvals, bvecs
    with open(bvals_filename.__str__(),'w') as f:
        f.writelines(bvals)
    with open(bvecs_filename.__str__(),'w') as f:
        f.writelines(bvecs)


def _write_dwi(filename,image , dest_type='nrrd',dtype='short'): ## image : image object (prep.dwi.DWI)
    if dest_type.lower()=='nrrd': ## load nrrd dwi image
        return _write_nrrd(image,filename,dtype=dtype)
    elif dest_type.lower()=='nifti':
        return _write_nifti(image, filename,dtype=dtype)
    else:
        logger("Not a supported image type",prep.Color.ERROR)
        raise Exception("Not a supported image type")
        return False
    
class DWI:
    def __init__(self,filename,b0_threshold=10,filetype=None,**kwargs):
        ## file information
        self.filename=filename
        
        ## Processed data 
        self.images=None #image tensors [ size x, size y, size z , gradient index]
        self.gradients=None #gradient {'index': , 'gradient' : }
        self.information=None #other image information such as b value , origin, ...
        self.b0_threshold=b0_threshold
        
        ## image specific data (not to be used in computation)
        self.image_type=filetype
        self.original_data=None #Original Data returned from each file type
        
        ## load image
        self.loadImage(self.filename,self.image_type)
        
    def __getitem__(self,index):
        return self.images[index,:,:,:], self.gradients[index]
    def __len__(self):
        return len(self.gradients)
    
    @staticmethod
    def mergeImages(*imgs):
        merged=copy.deepcopy(imgs[0])
        num_grads = merged.images.shape[-1]
        for idx,img in enumerate(imgs):
            if idx>0:
                temp_ngrads = img.images.shape[-1]
                num_grads = num_grads + temp_ngrads
                merged.images = np.concatenate((merged.images,img.images),axis=-1)
                merged.gradients = merged.gradients+img.gradients
        logger('Images merged',prep.Color.OK)
        return merged

    def setImage(self,img, modality='DWMRI', kinds = ['space','space','space','list']):
        self.information['sizes'] = list(img.shape)
        self.information['kinds'] = kinds
        self.information['original_kinds'] = kinds
        self.information['dimension'] = len(img.shape)
        self.information['modality'] = modality
        self.images = img
        self.gradients=[]

    @prep.measure_time
    def writeImage(self,filename,dest_type=None,dtype='short'):
        if not dest_type:
            if '.nrrd' in filename.lower(): 
                dest_type='nrrd'
            if '.nii' in filename.lower(): 
                dest_type='nifti'
        
        logger("Writing image {} to : {}".format(dest_type,str(filename)),prep.Color.PROCESS)
        _write_dwi(filename,self,dest_type=dest_type,dtype=dtype)
        logger("Image written.",prep.Color.OK)

    @prep.measure_time
    def loadImage(self,filename,filetype=None):
        print(self.filename)
        if '.nrrd' in filename.lower(): self.image_type='nrrd'
        if '.nii' in filename.lower(): self.image_type='nifti'
        if filetype is not None:
            self.image_type=filetype
        self.images,self.gradients,self.information ,self.original_data = _load_dwi(filename,self.image_type)
        self.images = self.images.astype(float)
        logger("Image - {} loaded".format(self.filename),prep.Color.OK,terminal_only=True)


    def getAffineMatrix(self):
        affine=np.transpose(np.append(self.information['space_directions'],
                                     np.expand_dims(self.information['space_origin'],0),
                                     axis=0))
        affine=np.append(affine,np.array([[0,0,0,1]]),axis=0)
        return affine 

    def getAffineMatrixForNifti(self):
        return self.getAffineMatrixBySpace('right-anterior-superior')

    def getAffineMatrixBySpace(self,target_space="right-anterior-superior"): #target_space left/right, posterior/anterior, inferior/superior e.g. lef-posterior-superior
        space=self.information['space']
        space_directions=copy.deepcopy(self.information['space_directions'])
        affine=copy.deepcopy(np.array(space_directions))
        # affine=self.getAffineMatrix()
        space_origin=np.array(self.information['space_origin'])
        affine=np.append(affine,[space_origin],axis=0)
        affine=affine.transpose()
        affine=np.append(affine,np.array([[0,0, 0, 1]]),axis=0)
        src_space_elem = space.split('-')
        target_space_elem = target_space.split('-')
        diag_elements = [1,1,1,1]
        for i,v in enumerate(target_space_elem):
            if v != src_space_elem[i]:
                diag_elements[i]=-1
        ijk_to_lps = affine
        src_to_tgt = np.diag(diag_elements)
        ijk_to_ras = np.matmul(src_to_tgt, ijk_to_lps)
        affine=ijk_to_ras
        return affine

    def getAffineMatrixForSlice(self,column=2): # 0 for x, 1 for y , 2 for z  output 2d affine matrix (3x3)
        affine=self.getAffineMatrix()
        new_mat=np.delete(affine,column,axis=0)
        new_mat=np.transpose(np.delete(np.transpose(new_mat),column,axis=0))
        return new_mat

    def setSpaceDirection(self, target_space=None):
        if not target_space:
            return
        affine = self.getAffineMatrixBySpace(target_space=target_space)
        at=affine.transpose()
        self.information['space_directions']=at[:3,:3].tolist()
        self.information['space_origin']=at[3,:3].tolist()
        self.information['space']=target_space

    def setB0Threshold(self,b0_threshold):
        self.b0_threshold=b0_threshold

    def getB0Threshold(self):
        return self.b0_threshold

    def getB0Index(self):
        grads=self.getGradients()
        b0threshold=self.getB0Threshold()
        res=[]
        for idx,g in enumerate(grads):
            if g['b_value'] <= b0threshold:
                res.append(idx)
        return res 

    def setGradients(self,gradients:list):
        _,_,_,g = self.images.shape 
        if g != len(gradients):
            logger("[ERROR] Gradients in the image doesn't match to the direction number of the gradient file",prep.Color.ERROR)
            logger("Number of gradients from image file : {}, Number of gradients from gradient file : {}".format(g,len(gradients)),prep.Color.ERROR)
            raise Exception("Gradients in the image doesn't match to the direction number of the gradient file")
        else:
            self.gradients=gradients 

    def getGradients(self,b0_threshold=None):
        if b0_threshold is None:
            b0_threshold=self.b0_threshold

        for idx,e in enumerate(self.gradients):
            self.gradients[idx]['index']=idx
            self.gradients[idx]['baseline']=bool(int(e['b_value'])<=b0_threshold)

        return self.gradients

    def removeGradients(self):
        self.gradients=[]
        del self.information['modality']

    def loadGradients(self,filename):
        self.gradients=yaml.safe_load(open(filename,'r'))

    def loadImageInformation(self,filename):
        self.information=yaml.safe_load(open(filename,'r'))


    def dumpInformation(self,filename):
        info=self.information
        yaml.safe_dump(info,open(filename,'w'))       

    def dumpGradients(self,filename):
        grad=self.getGradients()
        out_grad=[]
        for g in grad:
            temp={'index': int(g['index']),
                 'original_index': int(g['original_index']),
                 'b_value' : float(g['b_value']),
                 'gradient': g['gradient'],
                 'unit_gradient': g['unit_gradient'],
                 "baseline" : bool(g['baseline'])
            }
            out_grad.append(temp)
        yaml.safe_dump(out_grad,open(filename,'w'))

    def isGradientBaseline(self,gradient_index:int):
        return self.getGradients()[gradient_index]['baseline']

    def getBValueBounds(self):
        grad=self.getGradients()
        min_b= float(np.min([x['b_value'] for x in grad]) )
        max_b= float(np.max([x['b_value'] for x in grad]) )
        return [min_b,max_b]

    def getBaselines(self,b0_threshold=None):
        grads=self.getGradients(b0_threshold)
        baseline_gradients=[x for x in grads if x['baseline']]
        baseline_indexes=[x['index'] for x in baseline_gradients]
        baseline_volumes=self.images[:,:,:,baseline_indexes]
        return baseline_gradients, baseline_volumes

    def extractBaselines(self,b0_threshold=None):
        new_image=copy.deepcopy(self)
        grads,vols=new_image.getBaselines(b0_threshold)
        new_image.images=vols 
        new_image.setGradients(grads)
        return new_image

    def idwi(self,b0_threshold=None):
        gradients=self.getGradients(b0_threshold)
        num_gradients=len(gradients)
        vol_product=np.ones(self.images.shape[:3],dtype=np.float64)
        for idx,g in enumerate(gradients):
            vol_product*= self.images[:,:,:,idx]
        result_volume=pow(vol_product,1.0/num_gradients)
        return result_volume 

    def directAverage(self):
        result_volume=self.images.mean(axis=3)
        return result_volume 

    def reduceTo3D(self,method="direct_average"):
        if method=='idwi':
            return self.idwi()
        else: 
            return self.directAverage()

    def zerorizeBaselines(self,b0_threshold):
        grads=self.getGradients(b0_threshold)
        for idx,g in enumerate(grads):
            if g['baseline']:
                grads[idx]['b_value']=0
        self.setGradients(grads)

    def zeroPad(self,pad_list:list): ## [0,2,0,0] means padding 2 at axis 1 
        x,y,z,g=self.images.shape
        xa,ya,za,ga=pad_list 
        new_image=np.zeros([x+xa,y+ya,z+za,g+ga],dtype=self.images.dtype)
        
        new_image[:x,:y,:z,:g]=self.images
        self.images=new_image 
        self.information['image_size']=[x+xa,y+ya,z+za]
        self.information['sizes']=[x+xa,y+ya,z+za,g+ga]

        if ga >0 : #if gradient volume is added
            for i in range(ga):
                grad=self.gradients[-1]
                self.gradients.append(grad)

    def gradientSummary(self):
        grads=self.getGradients()
        num_baselines=len([x for x in grads if x['baseline']])
        num_gradients=len(grads)
        res={
            "number_of_baselines": num_baselines,
            "number_of_gradients": num_gradients
        }
        return res
        

    def convertToOriginalGradientIndex(self,grad_indexes:list): # from actual index to original gradient index list
        out=[]
        grad=self.getGradients()
        for idx in grad_indexes:
            out.append(grad[idx]['original_index'])
        return out 

    def deleteGradientsByOriginalIndex(self, original_indexes: list): #remove gradiensts and delete images corresponding to those gradients, list of gradient indexes
        ## remove gradient slices
        remove_list=[]
        grad=self.getGradients()
        for idx,g in enumerate(grad):
            if g['original_index'] in original_indexes:
                remove_list.append(idx)
        self.deleteGradients(remove_list)

    def deleteGradients(self,remove_list: list): #remove gradiensts and delete images corresponding to those gradients, list of gradient indexes
        ## remove gradient slices
        self.images=np.delete(self.images,remove_list,3)
        self.gradients=list(filter(lambda x: x['index'] not in remove_list, self.gradients))
        for idx,g in enumerate(self.gradients):
            g['index']=idx 
        self.update_information()

    def insertGradient(self,gradient,image_volume,pos=-1):
        self.images=np.insert(self.images,pos,image_volume,axis=3)
        self.gradients.insert(pos,gradient)
        self.update_information()

    def update_information(self):
        ## here goes anything to update when there is any changes on self.images
        self.information['sizes']=list(self.images.shape )
        self.information['image_size']=list(self.images.shape[:3])
        self.getGradients()
        ## reindexing



        
        




from PyQt5.QtWidgets import * 
from PyQt5 import QtCore, QtGui
from PyQt5.QtGui import * 
from PyQt5.QtCore import * 

from PyQt5.QtCore import QObject
from PyQt5.QtCore import pyqtSignal as Signal


class GeneralCommunicate(QObject):

  call_UnsavedModif = Signal()

  def CallUnsavedModif(self):
    self.call_UnsavedModif.emit()

class General(QWidget):
  communicate = GeneralCommunicate()

  def __init__(self, protocol_template, protocol_yml):
    QWidget.__init__(self)
    self.protocol_template = protocol_template
    self.tab = QWidget()

    self.GeneralParameters()
    self.InitializeParameters(protocol_yml)


  def GeneralParameters(self):
    general_layout = QFormLayout()
    self.tab.setLayout(general_layout)
    
    no_output_image = QHBoxLayout()
    self.no_output_image_false = QRadioButton("Yes")
    no_output_image.addWidget(self.no_output_image_false)
    self.no_output_image_true = QRadioButton("No")
    no_output_image.addWidget(self.no_output_image_true)
    general_layout.addRow("Create Output Image", no_output_image)
    self.no_output_image_false.toggled.connect(self.communicate.CallUnsavedModif)
    self.no_output_image_true.toggled.connect(self.communicate.CallUnsavedModif)

    self.output_format = QComboBox()
    for ite in self.protocol_template["options"]["io"]["output_format"]["candidates"]:
      self.output_format.addItem(ite["caption"])
      self.output_format.setItemData(self.output_format.count()-1, ite["description"], QtCore.Qt.ToolTipRole)
    self.output_format.currentTextChanged.connect(self.GetOutputImageTypeIt)
    general_layout.addRow(self.protocol_template["options"]["io"]["output_format"]["caption"], self.output_format)
    self.output_format.currentTextChanged.connect(self.communicate.CallUnsavedModif)

    self.no_output_image_false.clicked.connect(lambda: self.output_format.setEnabled(True))
    self.no_output_image_true.clicked.connect(lambda: self.output_format.setEnabled(False))

    self.b0 = QDoubleSpinBox()
    self.b0.setAlignment(Qt.AlignRight)
    self.b0.setSingleStep(0.1)
    self.b0.setDecimals(3)
    self.b0.setStatusTip(self.protocol_template["options"]["io"]["baseline_threshold"]["description"])
    general_layout.addRow(self.protocol_template["options"]["io"]["baseline_threshold"]["caption"], self.b0)
    self.b0.valueChanged.connect(self.communicate.CallUnsavedModif)

  def GetOutputImageTypeIt(self, text): # general parameter 
    self.outputimagetype_it = 0
    for it in range(len(self.protocol_template["options"]["io"]["output_format"]["candidates"])):
      if self.protocol_template["options"]["io"]["output_format"]["candidates"][it]["caption"] == text:
        self.outputimagetype_it = it
  
  def InitializeParameters(self, protocol_yml):
    if protocol_yml["io"]["no_output_image"] == False:
      self.no_output_image_false.setChecked(True)
    else:
      self.no_output_image_true.setChecked(True)
    image_type = protocol_yml["io"]["output_format"]
    for ite in self.protocol_template["options"]["io"]["output_format"]["candidates"]:
      if ite["value"] == image_type:
        self.output_format.setCurrentText(ite["caption"])
        self.GetOutputImageTypeIt(ite["caption"])
    self.b0.setValue(protocol_yml["io"]["baseline_threshold"])


from PyQt5.QtWidgets import * 
from PyQt5 import QtCore, QtGui
from PyQt5.QtGui import * 
from PyQt5.QtCore import * 

from dtiplayground.ui.modules_details_communicate import ModulesDetailsCommunicate

class UtilHeader(QWidget):
  communicate = ModulesDetailsCommunicate()

  def __init__(self, protocol_template, utilheader_yml):
    QWidget.__init__(self)
    self.stack = QWidget()
    self.UtilHeaderStack(protocol_template, utilheader_yml)

  def UtilHeaderStack(self, protocol_template, utilheader_yml):

    ## Module
    self.tab_name = QLabel()

    ## Description
    for template_ite in range(len(protocol_template["options"]["execution"]["pipeline"]["candidates"])):
      if protocol_template["options"]["execution"]["pipeline"]["candidates"][template_ite]["value"] == "UTIL_Header":
        description_label = QLabel(protocol_template["options"]["execution"]["pipeline"]["candidates"][template_ite]["description"])
    description_label.setWordWrap(True)

    ## Options:
    options_groupbox = QGroupBox("Options:")
    options_layout = QGridLayout()
    options_groupbox.setLayout(options_layout)   
    # overwrite
    self.overwrite = QCheckBox("Overwrite")
    self.overwrite.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.overwrite, 0, 0)
    # skip
    self.skip = QCheckBox("Skip")
    self.skip.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.skip, 0, 1)
    # write image
    self.writeimage = QCheckBox("Write Image")
    self.writeimage.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.writeimage, 0, 2)
    
    ## Protocol:
    protocol_groupbox = QGroupBox("Protocol:")
    protocol_layout = QFormLayout()
    protocol_groupbox.setLayout(protocol_layout)
    # options
    self.options = QLineEdit()
    self.options.setAlignment(Qt.AlignRight)
    self.options.setStatusTip(utilheader_yml["protocol"]["options"]["description"])
    protocol_layout.addRow(utilheader_yml["protocol"]["options"]["caption"], self.options)
    self.options.textChanged.connect(self.GetParams)

    ## Layout
    layout_v = QVBoxLayout()
    layout_v.addWidget(self.tab_name)
    layout_v.addWidget(description_label)
    layout_v.addWidget(options_groupbox)
    layout_v.addWidget(protocol_groupbox)
    layout_v.addStretch(1)
    self.stack.setLayout(layout_v)

  def GetParams(self):
    params = [
      'View Header',
      'UTIL_Header', 
      {'options': {
        'overwrite': self.overwrite.isChecked(), 
        'skip': self.skip.isChecked(), 
        'write_image': self.writeimage.isChecked()
        },
      'protocol': {
        'options': self.options.text(), 
        }
      }
    ]
    self.communicate.SendParams(params)


from dtiplayground.dmri.common.tools.base import ExternalToolWrapper

class GreedyAtlas(ExternalToolWrapper):
    def __init__(self,binary_path):
        super().__init__(binary_path)

    def compute_deformation_fields(self,xml_file,parsed_file):
        arguments=[
            '-f',xml_file,
            '-o',parsed_file
        ]
        self.setArguments(arguments)
        return self.execute(arguments)


from PyQt5.QtWidgets import * 

from functools import partial
from dtiplayground.ui.softwarepaths import SoftwarePaths
from dtiplayground.ui.dmriprepUI import Widgets
import dtiplayground.config as config


class Window(QMainWindow):

  def __init__(self, protocol_template, args):
    super().__init__()
    self.setWindowTitle("DMRIPrep - " + config.INFO["dmriprep"]["version"])
    dmriprep = Widgets(self, protocol_template, args)
    self.setCentralWidget(dmriprep)

    # Exit application
    exitAct = QAction('&Exit', self)
    exitAct.setShortcut('Ctrl+Q')
    exitAct.setStatusTip('Exit application')
    exitAct.triggered.connect(qApp.quit)
    
    # Save protocol file
    saveAct = QAction('&Save', self)
    saveAct.setShortcut('Ctrl+S')
    saveAct.setStatusTip('Save protocol file')
    saveAct.triggered.connect(dmriprep.protocol_tab.SaveProtocol)

    # Save As
    saveasAct = QAction('&Save As', self)
    saveasAct.setStatusTip('Save protocol file as...')
    saveasAct.triggered.connect(dmriprep.protocol_tab.SaveAs)

    # Configurate software paths
    configSoftware = QAction('Softwares', self)
    configSoftware.setStatusTip('Set software paths')
    self.software_window = SoftwarePaths()
    configSoftware.triggered.connect(self.ShowSoftwareWindow)

    self.statusBar() 
    
    # Menu
    menubar = self.menuBar()
    fileMenu = menubar.addMenu('&File')
    fileMenu.addAction(saveAct)
    fileMenu.addAction(saveasAct)
    fileMenu.addAction(exitAct)
    configMenu = menubar.addMenu('Config')
    configMenu.addAction(configSoftware)
    
    self.show()

  def ShowSoftwareWindow(self): 
    self.software_window.show()




import numpy as np 
from dipy.align.imaffine import (transform_centers_of_mass,
                                 AffineMap,
                                 MutualInformationMetric,
                                 AffineRegistration)
import dipy.align 
from dipy.align.transforms import (TranslationTransform3D,
                                   RigidTransform3D,
                                   AffineTransform3D)
import dtiplayground.dmri.preprocessing as prep
import copy

logger=prep.logger.write

def baseline_average(image_obj, opt,
                      averageInterpolationMethod='linear-interpolation',
                      averageMethod='BaselineOptimized',
                      b0Threshold=10,
                      stopThreshold=0.02,
                      maxIterations=2):
    
    image=copy.deepcopy(image_obj)
    averaged_baseline_image=None #2d image
    output=None

    ## baseline gradients and volum extraction  
    baseline_grads, _=image_obj.getBaselines(b0_threshold=b0Threshold)
    logger("Finding baseline gradients : ",prep.Color.INFO)
    for idx,g in enumerate(baseline_grads):
        logger("Gradient.idx {:03d} Original.idx {:03d} Direction {} B-value {:.1f}"
            .format(g['index'],g['original_index'],g['gradient'],g['b_value']),prep.Color.INFO)

    ### Check availability 
    no_baseline=False
    only_one_baseline=False
    if len(baseline_grads)<2:
        if len(baseline_grads)<1: no_baseline=True
        else: only_one_baseline=True

    ### computation
    if no_baseline:
        logger("[WARNING] There was no baseline found",prep.Color.WARNING)
        b0Threshold = min(list(map(lambda x: x['b_value'], image.getGradients())))
        #return None,[]
        output=direct_average(image, averageInterpolationMethod, b0Threshold,stopThreshold)
    elif averageMethod=='DirectAverage' or only_one_baseline:
        if only_one_baseline: logger("Only one baseline was found, averaging method will be changed to DirectAverage",prep.Color.WARNING)
        output=direct_average(image, averageInterpolationMethod, b0Threshold,stopThreshold)
    elif averageMethod=='BaselineOptimized':    
        output=baseline_optimized_average(image, averageInterpolationMethod, b0Threshold,stopThreshold,maxIterations)
    elif averageMethod=='BSplineOptimized':
        logger("[WARNING] BSplineOptimized method is NOT implemented, averaging method will be changed to baseline optimized averaging",prep.Color.WARNING)
        output=baseline_optimized_average(image, averageInterpolationMethod, b0Threshold,stopThreshold,maxIterations)

    averaged_baseline_volume = output['averaged_baseline']
    output_gradient= output['output_baseline_gradient'] ## single gradient
    baseline_gradients= output['baseline_gradients']

    #post processing. To remove baselines from the original image and insert the averaged basline image. Re-indexing of gradients
    image.deleteGradientsByOriginalIndex([x['original_index'] for x in baseline_gradients])
    image.insertGradient(output_gradient,averaged_baseline_volume,pos=0)
    excluded_gradients_original_indexes=[x['original_index'] for x in baseline_gradients]

    return image , excluded_gradients_original_indexes

def default_output_gradient():
    return {
        "index" : None,
        "original_index" : -1,
        "gradient" : [0.0,0.0,0.0],
        "unit_gradient" : [0.0,0.0,0.0],
        "b_value" : 0,
        "baseline" : True 
    }

def direct_average(image_obj, averageInterpolationMethod , b0Threshold, stopThreshold):
    logger("Direct averaging on the baseline(s) ... ",prep.Color.PROCESS)
    baseline_grads, baseline_images=image_obj.getBaselines(b0_threshold=b0Threshold)
    out_gradient=default_output_gradient()
    output={"averaged_baseline" : np.mean(baseline_images,axis=3) ,
            "output_baseline_gradient" : out_gradient, 
            "baseline_gradients" : baseline_grads}
    logger("Direct averaging DONE ",prep.Color.OK)
    return output

def baseline_optimized_average(image_obj, averageInterpolationMethod , b0Threshold, stopThreshold, maxIterations=2):
    logger("Baseline Optimized averaging on baselines ... ",prep.Color.PROCESS)
    baseline_grads, baseline_images=image_obj.getBaselines(b0_threshold=b0Threshold)
    out_gradient=default_output_gradient()
    affine=image_obj.getAffineMatrixForNifti()
    static=np.mean(baseline_images,axis=3) #initial direct averaging 
    previous_static=copy.deepcopy(static)
    averaged_image=copy.deepcopy(static)
    moving_images=copy.deepcopy(baseline_images)
    x,y,z,g = moving_images.shape

    succeeded=False
    for i in range(maxIterations):
        temp_images=[]
        logger("Iteration {}/{}".format(i+1,maxIterations),prep.Color.PROCESS)
        for gidx in range(g):

            ## rigid 3d registration to the averaged image
            logger("Rigid registration {}/{}".format(gidx+1,g),prep.Color.PROCESS)
            moving=moving_images[:,:,:,gidx]
            transformed, out_affine = rigid_3d(static,moving,affine,affine,sampling_prop=0.1)
            temp_images.append(transformed)
  
        moving_images=np.moveaxis(np.array(temp_images),0,-1) ## replace existing moving images with registered images
        static=np.mean(moving_images,axis=3) ## re average transformed moving images
        error=computeErrorRatio(static,previous_static)
               
        if error < stopThreshold:
            succeeded=True
            averaged_image=static 
            logger("Error ratio : {:.4f} < tolerance level {:.4f}".format(error,stopThreshold),prep.Color.OK)
            break
        else:
            logger("Error ratio : {:.4f} > tolerance level {:.4f}".format(error,stopThreshold),prep.Color.INFO)
        previous_static=copy.deepcopy(static)

    if not succeeded:
        logger("[WARNING] BaselineOptimized averaging failed, so direct averaging will be performed",prep.Color.WARNING)
        return direct_average(image_obj, averageInterpolationMethod, b0Threshold,stopThreshold)

    output={"averaged_baseline" : averaged_image,
            "output_baseline_gradient" : out_gradient, 
            "baseline_gradients" : baseline_grads}
    logger("Baseline Optimized averaging DONE",prep.Color.OK)
    return output

def computeErrorRatio(static,moving):

    sq_diff=np.mean((static-moving)**2)
    ratio = np.sqrt(sq_diff)/np.mean(moving)
    return ratio

def bspline_optimized_average(image_obj, averageInterpolationMethod , b0Threshold, stopThreshold):
    prep.not_implemented()
    logger("BSplineOptimized averaging on baselines ... ",prep.Color.PROCESS)
    baseline_grads, baseline_images=image_obj.getBaselines(b0_threshold=b0Threshold)
    out_gradient=default_output_gradient()
    output={"averaged_baseline" : None ,
            "output_baseline_gradient" : out_gradient, 
            "baseline_gradients" : baseline_grads}
    logger("BSplineOptimized averaging DONE",prep.Color.OK)
    return output

def decompose_affine_matrix(mat4d): ## in case 
    scale, shear, angles, trans, persp =dipy.align.streamlinear.decompose_matrix(mat4d)
    angles_in_deg=list(map(lambda x : float(np.rad2deg(x)), angles))
    res={"scale":scale.tolist(),
         "shear":list(map(float,shear)),
         "angles":angles,
         "angles_in_degree":angles_in_deg,
         "translations":trans.tolist(),
         "perspective":persp.tolist()}
    return res

@prep.measure_time
def rigid_3d(static,moving,
             affine_static,affine_moving,
             nbins=32,
             level_iters=[10000,1000,100],
             sigmas=[3.0,1.0,0.0],
             factors=[4,2,1],
             sampling_prop=None):
    ## Make affine map
    identity=np.eye(4)
    affine_map=AffineMap(identity,static.shape, affine_static,
                                  moving.shape, affine_static)

    ## Resample moving image from affine (but same in a volume)
    resampled=affine_map.transform(moving)

    ## center of mass transform
    c_of_mass = transform_centers_of_mass(static, affine_static, moving, affine_moving)
    transformed=c_of_mass.transform(moving)

    ## registration preparation
    metric= MutualInformationMetric(nbins,sampling_prop)
    affreg= AffineRegistration(metric=metric, 
                               level_iters=level_iters, 
                               sigmas=sigmas, 
                               factors=factors,
                               verbosity=0)

    ## Translation transform registration
    transform=TranslationTransform3D()
    params0=None
    starting_affine= c_of_mass.affine
    translation=affreg.optimize(static, 
                                moving, 
                                transform, 
                                params0, 
                                affine_static,affine_moving, 
                                starting_affine=starting_affine)
    transformed = translation.transform(moving)

    ## Ridid registration
    transform=RigidTransform3D()
    params0=None
    starting_affine=translation.affine
    rigid=affreg.optimize(static,moving,transform,params0,affine_static,affine_moving,starting_affine=starting_affine)
    transformed = rigid.transform(moving)

    return transformed, rigid.affine 




from dtiplayground.dmri.common.tools.base import ExternalToolWrapper

class ImageMath(ExternalToolWrapper):
    def __init__(self,binary_path):
        super().__init__(binary_path)

    def rescale(self,inputfile, outputfile , rescale=[0,10000]):
        arguments=[inputfile,'-outfile',outfile,'-rescale',",".join(rescale)]
        self.setArguments(arguments)
        return self.execute(arguments)

    def normalize(self,inputfile,outputfile,referencefile):
        arguments=[inputfile,'-outfile',outputfile,'-matchHistogram',referencefile]
        self.setArguments(arguments)
        return self.execute(arguments)

    def average(self,inputfile,outputfile,files_to_average:list):
        arguments=[inputfile,'-outfile',outputfile,'-avg']+files_to_average
        self.setArguments(arguments)
        return self.execute(arguments)




    


from dtiplayground.dmri.common.tools.base import ExternalToolWrapper

class UNU(ExternalToolWrapper):
    def __init__(self,binary_path):
        super().__init__(binary_path)

    def convert_to_float(self,input_file,output_file):
        arguments=[
            'convert',
            '-t','float',
            '-i',input_file,
        ]
        self.setArguments(arguments)
        pipe_output=self.execute_pipe()
        arguments=[
            'save',
            '-f','nrrd',
            '-e','gzip',
            '-o',output_file
        ]
        self.setArguments(arguments)
        return self.execute(stdin=pipe_output.stdout)


from dtiplayground.dmri.common.tools.base import ExternalToolWrapper

class ResampleDTIlogEuclidean(ExternalToolWrapper):
    def __init__(self,binary_path):
        super().__init__(binary_path)

    def filter_dti(self, inputfile, outputfile,correction='zero'):
        arguments=[inputfile,outputfile,'--correction',correction]
        self.setArguments(arguments)
        return self.execute(arguments)

    def implement_affine_registration(self,input_file,affine_file,transform_file,reference_file):
        arguments=[input_file,affine_file,'-f',transform_file,'-R',reference_file]
        self.setArguments(arguments)
        return self.execute(arguments)

    def resample(self,reference_file,
                       deform_field_file,
                       transformation_file,
                       moving_file,
                       output_file,
                       interpolation_type='Linear',
                       interpolation_option=None,
                       tensor_interpolation_type='Non Log Euclidean',
                       tensor_interpolation_option='Zero',
                       tensor_transform='Preservation of the Principal Direction (PPD)'
                       ):
        arguments=[
            '-R',reference_file,
            '-H',deform_field_file,
            '-f',transformation_file,
            moving_file,
            output_file
        ]
        options=[]
        ## options
        if interpolation_type=="Linear" : options+=["-i","linear"]
        if interpolation_type=="Nearest Neighborhood" : options+=["-i","nn"]
        if interpolation_type=="Windowed Sinc":
          if interpolation_option=="Hamming": options+=["-i", "ws", "-W", "h"]
          if interpolation_option=="Cosine" : options+=["-i", "ws", "-W", "c"] 
          if interpolation_option=="Welch"  : options+=["-i", "ws", "-W", "w"] 
          if interpolation_option=="Lanczos": options+=["-i", "ws", "-W", "l"] 
          if interpolation_option=="Blackman":options+=["-i", "ws", "-W", "b"]
        if interpolation_type=="BSpline":
          options+=["-i","bs","-o"] + [m_interpolation_option]  
        if tensor_interpolation_type=="Non Log Euclidean":
          if tensor_interpolation_option=="Zero" :  options+=["--nolog","--correction","zero"] #FinalReSampCommand = FinalReSampCommand + " --nolog --correction zero"
          if tensor_interpolation_option=="None" :  options+=["--nolog","--correction","none"] #FinalReSampCommand = FinalReSampCommand + " --nolog --correction none"
          if tensor_interpolation_option=="Absolute Value" : options+=["--nolog","--correction","abs"] #FinalReSampCommand = FinalReSampCommand + " --nolog --correction abs"
          if tensor_interpolation_option=="Nearest" : options+=["--nolog","--correction","nearest"] #FinalReSampCommand = FinalReSampCommand + " --nolog --correction nearest"
        if tensor_transform=="Preservation of the Principal Direction (PPD)": options+=['-T','PPD']
        if tensor_transform=="Finite Strain (FS)" :  options+=['-T','FS']

        arguments+=options
        self.setArguments(arguments)
        return self.execute(arguments)


from dtiplayground.dmri.common.tools.base import ExternalToolWrapper

class DTIProcess(ExternalToolWrapper):
    def __init__(self,binary_path):
        super().__init__(binary_path)

    
    def measure_scalars(self,inputfile,outputfile,scalar_type='FA',options=[]):
        assert(scalar_type.lower() in ['fa','md'])
        scalar_opt='-f' #FA
        if scalar_type.lower()!='fa' : scalar_opt='-m' #MA
        arguments=['--dti_image',inputfile,scalar_opt,outputfile]+options
        self.setArguments(arguments)
        return self.execute(arguments)




#
# Reference : https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3864968/
#
# Written by SK Park , NIRAL, UNC
# 2021-04-18

import dtiplayground.dmri.preprocessing as prep

import yaml,traceback 
import BASELINE_Average.computations as computations 
import numpy as np
import time
from pathlib import Path

logger=prep.logger.write


class BASELINE_Average(prep.modules.DTIPrepModule):
    def __init__(self,config_dir,*args,**kwargs):
        super().__init__(config_dir,*args,**kwargs)
        
    def generateDefaultProtocol(self,image_obj):
        super().generateDefaultProtocol(image_obj)
        ## todos
        return self.protocol

        
    def process(self,*args,**kwargs): ## variables : self.source_image, self.image (output) , self.result_history , self.result (output) , self.protocol, self.template
        super().process()
        inputParams=self.getPreviousResult()['output']
        opts=args[0]
        self.baseline_threshold=opts['baseline_threshold']

        output=None
        output_image_path=Path(self.output_dir)
        if self.image.image_type.lower()=='nrrd':
            output_image_path=str(output_image_path.joinpath('output.nrrd'))
        elif self.image.image_type.lower()=='nifti':
            output_image_path=str(output_image_path.joinpath('output.nii.gz'))
        
        output_filename=Path(self.computation_dir).joinpath('computations.yml')
        if output_filename.exists() and not self.options['recompute']: 
            ## pass recomputation
            logger("Computing ommited",prep.Color.INFO)
            pass
        else: ##computed parameters doesn't exist or recompute is true
            ## compute or recompute
            logger("Computing ... ",prep.Color.PROCESS)
            #self.image.deleteGradientsByOriginalIndex([49, 65, 97, 129, 145])#([0, 17, 49, 65, 97, 129, 145]) For test
            new_image, excluded_original_indexes=computations.baseline_average(self.image, opt=None ,
                                                      averageInterpolationMethod=self.protocol['averageInterpolationMethod'],
                                                      averageMethod=self.protocol['averageMethod'],
                                                      b0Threshold=self.baseline_threshold,
                                                      stopThreshold=self.protocol['stopThreshold'],
                                                      maxIterations=self.protocol['maxIterations'])

        if new_image is not None:
            self.image=new_image
            ### if image is changed, next module should load the file. So set image_object to None and write the file instead
        self.image.setSpaceDirection(target_space=self.getSourceImageInformation()['space'])
        self.writeImage(output_image_path,dest_type=self.image.image_type)

        ### output preparation
        self.result['output']['excluded_gradients_original_indexes']=excluded_original_indexes
        self.result['output']['success']=True
        #raise Exception("User Exception for development ...")
        return self.result



  
import dtiplayground.dmri.preprocessing as prep
import yaml, os
from pathlib import Path
logger=prep.logger.write

import dtiplayground.dmri.common.tools as tools 

class DTI_Register(prep.modules.DTIPrepModule):
    def __init__(self,config_dir,*args,**kwargs):
        super().__init__(config_dir,*args,**kwargs)
        
    def generateDefaultProtocol(self,image_obj):
        super().generateDefaultProtocol(image_obj)
        self.protocol['ANTsPath'] = self.protocol['ANTsPath'].replace('$ANTSDIR', self.softwares['ANTs']['path'])
        ## todos
        if 'reference_dti' in self.global_variables:
            ref_dti_path = Path(self.global_variables['reference_dti'])
            self.protocol['referenceImage'] = ref_dti_path.resolve().__str__()
        return self.protocol

    def process(self,*args,**kwargs): ## variables : self.global_variables, self.softwares, self.output_dir, self.config_dir, self.source_image, self.image (output) , self.result_history , self.result (output) , self.protocol, self.template
        super().process()
        inputParams=self.getPreviousResult()['output']
        protocol_options=args[0]
        self.num_threads=protocol_options['software_info']['parameters']['num_max_threads']
        self.baseline_threshold=protocol_options['baseline_threshold']

        # << TODOS>>
        self.dtiImagePath = None
        if 'dti_image' in self.global_variables:
            self.dtiImagePath=self.global_variables['dti_image']
        self.register(**self.protocol)

        logger(yaml.dump(self.image.information))
        self.result['output']['success']=True

        return self.result

    def register(self,**protocol):
        if protocol['method'] == 'ANTs':
            logger("ANTS is selected for the method",prep.Color.INFO)
            self.registerWithANTs(**protocol)
        else:
            logger("No method was selected, skipping the registration",prep.Color.WARNING)
        return True

    def registerWithANTs(self,**protocol):
        output_dir = Path(self.output_dir)
        refImagePath = self.protocol['referenceImage']
        if not Path(refImagePath).exists():
            logger("Reference image doesn't exist. Please check the file {} exists.".format(refImagePath),prep.Color.ERROR)
            raise Exception("File not found")
        inputImagePath = output_dir.joinpath('input.nrrd').__str__()
        registeredImagePath = output_dir.joinpath('registered.nrrd').__str__()
        # outputImagePath = output_dir.joinpath('output.nrrd').__str__()
        displacementFieldPath = output_dir.joinpath('displacementField.nrrd').__str__()
        inv_displacementFieldPath = output_dir.joinpath('inverse_displacementField.nrrd').__str__()
        outputDirectory = output_dir.__str__()
        nbThreads = self.num_threads

        ANTsMethod = self.protocol['ANTsMethod']
        registrationType = self.protocol['registrationType']
        similarityMetric = self.protocol['similarityMetric']
        similarityParameter = self.protocol['similarityParameter']
        ANTSIterations = self.protocol['ANTsIterations']
        gaussianSigma = self.protocol['gaussianSigma']
        
        ## saving input
        if self.dtiImagePath is not None:
            inputImagePath = self.dtiImagePath
        else:
            self.writeImageWithOriginalSpace(inputImagePath,'nrrd',dtype='float')

        os.environ['ITK_GLOBAL_DEFAULT_NUMBER_OF_THREADS'] = str(nbThreads)
        ANTsPath=Path(protocol['ANTsPath']).joinpath('bin/ANTS').__str__()
        logger("Executing DTI Reg for registration",prep.Color.PROCESS)
        args = ['--fixedVolume',refImagePath,
                '--movingVolume',inputImagePath,
                '--method',ANTsMethod,
                '--ANTSRegistrationType',registrationType,
                '--ANTSSimilarityMetric',similarityMetric,
                '--ANTSSimilarityParameter',str(similarityParameter),
                '--ANTSGaussianSigma',str(gaussianSigma),
                '--ANTSIterations',ANTSIterations,
                '--outputDisplacementField',displacementFieldPath,
                '--outputInverseDeformationFieldVolume', inv_displacementFieldPath,
                '--ANTSPath',ANTsPath,
                '--dtiprocessPath',self.softwares['dtiprocess']['path'],
                '--ResampleDTIPath',self.softwares['ResampleDTIlogEuclidean']['path'],
                '--ITKTransformToolsPath',self.softwares['ITKTransformTools']['path'],
                '--outputVolume', registeredImagePath]
                # '--outputFolder',outputDirectory]
        dtireg=tools.DTIReg(self.softwares['DTI-Reg']['path'])
        dtireg.dev_mode=True
        dtireg.execute_with_args(args)
        self.result['output']['displacement_field_path'] = displacementFieldPath
        self.result['output']['inverse_displacement_field_path'] = inv_displacementFieldPath
        self.result['output']['registered_dti_image'] = registeredImagePath
        self.addGlobalVariable('displacement_field_path',displacementFieldPath)
        self.addGlobalVariable('inverse_displacement_field_path',inv_displacementFieldPath)
        self.addGlobalVariable('registered_dti_image',registeredImagePath)
        self.addGlobalVariable('reference_dti_image',refImagePath)

        # if self.protocol['useRegistered']:
        #    self.loadImage(registeredImagePath) 
        self.addOutputFile(registeredImagePath, 'DTI_Registered')
        self.addOutputFile(displacementFieldPath, 'DTI_DisplacementField')
        self.addOutputFile(inv_displacementFieldPath, 'DTI_Inverse_DisplacementField')

        #self.writeImageWithOriginalSpace(outputImagePath,'nrrd',dtype='float')
        return True 


import yaml
from pathlib import Path
import pandas
import os
import fnmatch
import SimpleITK as sitk
import numpy
from PIL import Image
from reportlab.pdfgen import canvas
from reportlab.lib.pagesizes import letter
from PyPDF2 import PdfFileMerger

import dtiplayground.dmri.preprocessing as prep

logger=prep.logger.write

class QC_Report(prep.modules.DTIPrepModule):
    def __init__(self,config_dir,*args,**kwargs):
        super().__init__(config_dir,*args,**kwargs)
        
    def generateDefaultProtocol(self,image_obj):
        super().generateDefaultProtocol(image_obj)
        ## todos
        return self.protocol

    def process(self,*args,**kwargs): ## variables : self.config_dir, self.source_image, self.image (output) , self.result_history , self.result (output) , self.protocol, self.template
        super().process()
        inputParams=self.getPreviousResult()['output']
        # << TODOS>>
        data = self.GetData()
        if self.protocol["generateCSV"] == True:
            self.CreateCSV(data)
        if self.protocol["generatePDF"] == True:
            info_display_QCed_gradients = self.CreateImages()
            self.CreatePDF(data, info_display_QCed_gradients)

        self.result['output']['success']=True
        return self.result

    def GetData(self):
        data = []

        # SINGLE INPUT
        if type(self.result_history[0]["output"]) != list:
            input_image = self.result_history[0]["output"]["image_path"]
            image_data = {'input_image': input_image, 'modules':{}}
            total_excluded_gradients = []     
            number_of_modules = (len(self.result_history) - 1)
            for i in range(1, number_of_modules + 1):
                image_data['modules'][self.result_history[i]["module_name"]] =[]
                # get excluded gradients
                if self.result_history[i]["module_name"] in ["SLICE_Check", "INTERLACE_Check", "MANUAL_Exclude", "EDDYMOTION_Correct"]:
                    excluded_gradients = self.result_history[i]["output"]["excluded_gradients_original_indexes"]
                    image_data['modules'][self.result_history[i]["module_name"]] = excluded_gradients
                    total_excluded_gradients += excluded_gradients
                # get rms 
                if self.result_history[i]["module_name"] == "EDDYMOTION_Correct":
                    path_rms = str(Path(self.output_dir).parent.parent) + "/" + self.result_history[i]["output"]["output_directory"] + "/output_eddied.eddy_movement_rms"
                    data_rms = pandas.read_csv(path_rms, sep = '  ', engine = 'python', usecols = [1])
                    rmsLargerThan1 = data_rms[data_rms > 1.0].count()[0]
                    rmsLargerThan2 = data_rms[data_rms > 2.0].count()[0]
                    rmsLargerThan3 = data_rms[data_rms > 3.0].count()[0]
                    image_data["rms"] = {"largerThan1": rmsLargerThan1, "largerThan2": rmsLargerThan2, "largerThan3": rmsLargerThan3}
            # get original number of gradients
            for number in self.result_history[0]["output"]["image_information"]["sizes"]:
                if number not in self.result_history[0]["output"]["image_information"]["image_size"]:
                    image_data["original_number_of_gradients"] = number
            image_data["number_of_excluded_gradients"] = len(total_excluded_gradients)
            data.append(image_data)
        

        # MULTI INPUT
        else:
            input_images_directory_list = [] 
            for image in self.result_history[0]["output"]:
                if "input" in image:
                    input_images_directory_list.append(str(Path(image["input"]["output_directory"]).parent))
            for input_image in input_images_directory_list:
                image_directory_path = str(Path(self.output_dir).parent.parent) + "/" + input_image
                list_modules_directory = []
                for element in os.scandir(image_directory_path):
                    if element.is_dir():
                        list_modules_directory.append(element.path)
                list_modules_result = []
                for module_directory in list_modules_directory:
                    list_modules_result.append(yaml.safe_load(open(module_directory + "/result.yml", 'r')))
                image_data = {'input_image': list_modules_result[0]["input"]["image_path"], 'modules':{}}
                total_excluded_gradients = []
                number_of_modules = len(list_modules_directory)
                for module_result in list_modules_result:
                    image_data['modules'][module_result["module_name"]] = []
                    # get excluded gradients
                    if module_result["module_name"] in ["SLICE_Check", "INTERLACE_Check", "MANUAL_Exclude", "EDDYMOTION_Correct"]:
                        excluded_gradients = module_result["output"]["excluded_gradients_original_indexes"]
                        image_data['modules'][module_result["module_name"]] = excluded_gradients
                        total_excluded_gradients += excluded_gradients
                # get original number of gradients
                for number in list_modules_result[0]["input"]["image_information"]["sizes"]:
                    if number not in list_modules_result[0]["input"]["image_information"]["image_size"]:
                        image_data["original_number_of_gradients"] = number
                image_data["number_of_excluded_gradients"] = len(total_excluded_gradients)
                data.append(image_data)

            if len(input_images_directory_list) == 0: #no module before fusionning images
                for input_image in self.result_history[0]["output"]:
                    image_data = {'input_image': input_image["output"]["image_path"], 'modules':{}}
                    image_data["number_of_excluded_gradients"] = 0
                    # get original number of gradients
                    for number in input_image["output"]["image_information"]["sizes"]:
                        if number not in input_image["output"]["image_information"]["image_size"]:
                            image_data["original_number_of_gradients"] = number
                    data.append(image_data)

            input_image = self.result_history[1]["output"]["image_path"]
            image_data = {'input_image': input_image, 'modules':{}}
            total_excluded_gradients = []     
            number_of_modules = len(self.result_history) - 1
            for i in range(1, number_of_modules + 1):
                image_data['modules'][self.result_history[i]["module_name"]] =[]
                # get excluded gradients
                if self.result_history[i]["module_name"] in ["SLICE_Check", "INTERLACE_Check", "MANUAL_Exclude", "EDDYMOTION_Correct"]:
                    excluded_gradients = self.result_history[i]["output"]["excluded_gradients_original_indexes"]
                    image_data['modules'][self.result_history[i]["module_name"]] = excluded_gradients
                    total_excluded_gradients += excluded_gradients
                # get rms 
                if self.result_history[i]["module_name"] == "EDDYMOTION_Correct":
                    path_rms = str(Path(self.output_dir).parent.parent) + "/" + self.result_history[i]["output"]["output_directory"] + "/output_eddied.eddy_movement_rms"
                    data_rms = pandas.read_csv(path_rms, sep = '  ', engine = 'python', usecols = [1])
                    rmsLargerThan1 = data_rms[data_rms > 1.0].count()[0]
                    rmsLargerThan2 = data_rms[data_rms > 2.0].count()[0]
                    rmsLargerThan3 = data_rms[data_rms > 3.0].count()[0]
                    image_data["rms"] = {"largerThan1": rmsLargerThan1, "largerThan2": rmsLargerThan2, "largerThan3": rmsLargerThan3}
            # get original number of gradients
            image_data["original_number_of_gradients"] = 0
            image_data["number_of_excluded_gradients"] = len(total_excluded_gradients)
            for input_image_index in data:
                image_data["original_number_of_gradients"] += input_image_index["original_number_of_gradients"]
                image_data["number_of_excluded_gradients"] += input_image_index["number_of_excluded_gradients"]
            data.append(image_data)
        
        return(data)

    ## CSV

    def CreateCSV(self, data):
        if len(data) == 1:
            columns = ["image_name", 'original_number_of_gradients', 'number_of_excluded_gradients']
            values = [data[0]["input_image"], data[0]["original_number_of_gradients"], data[0]["number_of_excluded_gradients"]]
            if "rms" in data[0]:
                columns += ['rms_larger_than_1', 'rms_larger_than_2', 'rms_larger_than_3']
                values += [data[0]["rms"]["largerThan1"], data[0]["rms"]["largerThan2"], data[0]["rms"]["largerThan3"]]
        if len(data) > 1: #3 images due to susceptibility correction module
            columns = ["image_name_1", 'original_number_of_gradients_1', 'number_of_excluded_gradients_1']
            values = [data[0]["input_image"], data[0]["original_number_of_gradients"], data[0]["number_of_excluded_gradients"]]
            if "rms" in data[0]:
                columns += ['rms_larger_than_1_image1', 'rms_larger_than_2_image1', 'rms_larger_than_3_image1']
                values += [data[0]["rms"]["largerThan1"], data[0]["rms"]["largerThan2"], data[0]["rms"]["largerThan3"]]
            columns += ["image_name_2", 'original_number_of_gradients_2', 'number_of_excluded_gradients_2']
            values += [data[1]["input_image"], data[1]["original_number_of_gradients"], data[1]["number_of_excluded_gradients"]]
            if "rms" in data[1]:
                columns += ['rms_larger_than_1_image2', 'rms_larger_than_2_image2', 'rms_larger_than_3_image2']
                values += [data[1]["rms"]["largerThan1"], data[1]["rms"]["largerThan2"], data[1]["rms"]["largerThan3"]]
            columns += ["image_name_combined", 'original_number_of_gradients_combined', 'number_of_excluded_gradients_combined']
            values += [data[2]["input_image"], data[2]["original_number_of_gradients"], data[2]["number_of_excluded_gradients"]]
            if "rms" in data[2]:
                columns += ['rms_larger_than_1_combined', 'rms_larger_than_2_combined', 'rms_larger_than_3_combined']
                values += [data[2]["rms"]["largerThan1"], data[2]["rms"]["largerThan2"], data[2]["rms"]["largerThan3"]]

        qc_report = pandas.DataFrame([values], columns = columns)
        path_output_directory = Path(self.output_dir).parent.parent
        qc_report.to_csv(str(path_output_directory) + "/qc_report.csv", index=False)

    ## PDF

    def CreatePDF(self, data, info_display_QCed_gradients):
        temp_file = False  
        eddymotion_in_protocol = False
        for image in data:
            for module in image["modules"].keys():
                if module == "EDDYMOTION_Correct":
                    eddymotion_in_protocol = True
        if eddymotion_in_protocol:
            eddy_folder_path = Path(self.output_dir).parent
            for dirname in os.listdir(eddy_folder_path):
                if fnmatch.fnmatch(dirname, "*_EDDYMOTION_Correct"):
                    eddy_report_path = str(eddy_folder_path) + "/" + dirname + "/output_eddied.qc/qc.pdf"
                    eddy_report_path_abs = os.path.abspath(eddy_report_path)
                    print(eddy_report_path_abs)
                    if os.path.isfile(eddy_report_path_abs):
                        temp_file = True
                        print("file exists !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!1")


        if temp_file:
            pdf_file = self.output_dir + '/temp_QC_Report.pdf'
        else:
            pdf_file = str(Path(self.output_dir).parent.parent) + '/QC_Report.pdf'

        # sumup
        can = canvas.Canvas(pdf_file, pagesize=letter)
        x, y = 20, 750
        dx, dy = 15, 20
        for image in data:
            text = "Image: " + image["input_image"]
            while len(text) % 90 > 0:
                can.drawString(x, y, text[:90])
                text = text[90:]
                y -=10                
            can.drawString(x, y, text)
            y -= dy
            for module in image["modules"]:
                can.drawString(x+dx, y, module)
                y -= dy
                if module in ["SLICE_Check", "INTERLACE_Check", "MANUAL_Exclude"]:
                    excluded_gradients = image["modules"][module]
                    if len(excluded_gradients) == 0:
                        can.drawString(x+2*dx, y, "0 exluded gradient")
                        y -= dy
                    else:
                        excluded_gradients_text = str(excluded_gradients[0]) 
                        for index in range(1, len(excluded_gradients)):
                            excluded_gradients_text = excluded_gradients_text + ", " + str(excluded_gradients[index])
                        text = str(len(excluded_gradients)) + " excluded gradient(s): " + excluded_gradients_text
                        while len(text) % 90 > 0:
                            can.drawString(x+2*dx, y, text[:90])
                            text = text[90:]
                            y -=10              
                        can.drawString(x+2*dx, y, text)
                        y -= dy
                if module == "EDDYMOTION_Correct":
                    can.drawString(x+2*dx, y, str(image["rms"]["largerThan1"]) + " gradients with RMS movement relative to first volume > 1 mm")
                    y -= dy
                    can.drawString(x+2*dx, y, str(image["rms"]["largerThan2"]) + " gradients with RMS movement relative to first volume > 2 mm")
                    y -= dy
                    can.drawString(x+2*dx, y, str(image["rms"]["largerThan3"]) + " gradients with RMS movement relative to first volume > 3 mm")
                    y -= dy
            can.drawString(x+dx, y, "Total: " + str(image["number_of_excluded_gradients"]) + " gradient(s) excluded out of " + str(image["original_number_of_gradients"]))
            y -= dy
            can.drawString(x+dx, y, str(100 - round(image["number_of_excluded_gradients"]/image["original_number_of_gradients"], 2)*100) + "% of original gradients are preserved")
            y -= 2*dy
        can.showPage()

        # QCed volume gradients
        x, y = 20, 750 
        output_number_gradients = info_display_QCed_gradients[0]
        width = info_display_QCed_gradients[1]
        height = info_display_QCed_gradients[2]
        for dwi in range(output_number_gradients):
            image_path = self.output_dir + "/QC_Report_images/dwi" + str(dwi) + ".jpg"
            if dwi % 2 == 0:
                x = (4*612 - 3*width) // 12
            else:
                x = (8*612 + width) // 12
            can.drawString(x, y, "DWI " + str(dwi))
            if dwi % 2 == 0:
                x = (612 - 2*width) // 3
            else:
                x = (2*612 - width) // 3
            can.drawImage(image_path, x, y-height-dy//8, width = width, height = height)
            if dwi % 2 != 0:
                y = y - height - dy
            if y - height <= 45:
                y = 750
                can.showPage()
            x = 20
        can.save()

        if temp_file:
            self.MergePDF(pdf_file, eddy_report_path) #add figures from output_eddied.qc/qc.pdf (fsl)

    def MergePDF(self, temp_PDF_path, eddy_report_path):
        pdf_merger = PdfFileMerger()
        pdf_merger.append(temp_PDF_path)
        pdf_merger.merge(1, str(eddy_report_path), pages=(2, 3))
        pdf_merger.merge(2, str(eddy_report_path), pages=(4, 5))
        pdf_merger.merge(3, str(eddy_report_path), pages=(5, 6))
        pdf_merger.write(str(Path(self.output_dir).parent.parent) + '/QC_Report.pdf')




    ## Images

    def CreateImages(self):
        
        input_image = sitk.GetImageFromArray(self.source_image.images)
        input_size = list(input_image.GetSize())
        input_number_gradients = list(self.source_image.images.shape)[3]
        
        output_images_directory = self.GetOutputImagesDirectory()
        zoom_factor = self.GetZoom(input_size, input_image)

        for iter_gradients in range(input_number_gradients):  

            axial_image = self.AxialView(zoom_factor, iter_gradients, input_size, input_image)
            sagittal_image = self.SagittalView(zoom_factor, iter_gradients, input_size, input_image)
            coronal_image = self.CoronalView(zoom_factor, iter_gradients, input_size, input_image)

            # concatenate
            width = axial_image.height + sagittal_image.width + coronal_image.height
            height = max(axial_image.width, sagittal_image.height, coronal_image.width)
            dwi_image = Image.new('L', (width, height), 0)
            dwi_image.paste(sagittal_image, (0, 0))
            dwi_image.paste(axial_image, (sagittal_image.width, 0))
            dwi_image.paste(coronal_image, (sagittal_image.width + axial_image.height, 0))
            dwi_image.save(output_images_directory + "/dwi" + str(iter_gradients) + ".jpg")
  
        info_display_QCed_gradients = [input_number_gradients, dwi_image.width, dwi_image.height]
        return info_display_QCed_gradients

    def GetOutputImagesDirectory(self):
        if not os.path.exists(self.output_dir + "/QC_Report_images"):
            os.mkdir(self.output_dir + "/QC_Report_images")
        return str(self.output_dir) + "/QC_Report_images"

    
    def GetZoom(self, input_size, input_image):
        
        axial_image = self.AxialView(1, 0, input_size, input_image)
        sagittal_image = self.SagittalView(1, 0, input_size, input_image)
        coronal_image = self.CoronalView(1, 0, input_size, input_image)
        width = axial_image.height + sagittal_image.width + coronal_image.height
        zoom_factor = round((612 -3*20)/(2*width), 2)
        return zoom_factor


    def AxialView(self, zoom_factor, iter_gradients, input_size, input_image):

        slice_extractor = sitk.ExtractImageFilter()  
        slice_extractor.SetSize([input_size[0], input_size[1], 0])
        slice_extractor.SetIndex([0, 0, input_size[2]//2])
        extracted_slice = slice_extractor.Execute(input_image)

        gradient_extractor = sitk.VectorIndexSelectionCastImageFilter()
        gradient_extractor.SetIndex(iter_gradients)
        gradient = gradient_extractor.Execute(extracted_slice)
        
        gradient_array = sitk.GetArrayFromImage(gradient)
        gradient_array_normalized = (gradient_array - numpy.min(gradient_array)) * round(255 / numpy.max(gradient_array), 3)
        gradient_image = Image.fromarray(gradient_array_normalized)
        gradient_image = gradient_image.convert("L")
        gradient_image = gradient_image.rotate(90)
        gradient_image_resized = gradient_image.resize((int(zoom_factor*gradient_image.width), int(zoom_factor*gradient_image.height)))
        return gradient_image_resized

    def SagittalView(self, zoom_factor, iter_gradients, input_size, input_image):
        
        slice_extractor = sitk.ExtractImageFilter()
        slice_extractor.SetSize([0, input_size[1], input_size[2]])
        slice_extractor.SetIndex([input_size[0]//2, 0, 0])
        extracted_slice = slice_extractor.Execute(input_image)

        gradient_extractor = sitk.VectorIndexSelectionCastImageFilter()
        gradient_extractor.SetIndex(iter_gradients)
        gradient = gradient_extractor.Execute(extracted_slice)
        
        gradient_array = sitk.GetArrayFromImage(gradient)
        gradient_array_normalized = (gradient_array - numpy.min(gradient_array)) * round(255 / numpy.max(gradient_array), 3)
        gradient_array_normalized = numpy.flipud(gradient_array_normalized)
        gradient_image = Image.fromarray(gradient_array_normalized)
        gradient_image = gradient_image.convert("L")
        gradient_image = gradient_image.rotate(270)
        gradient_image_resized = gradient_image.resize((int(zoom_factor*gradient_image.width), int(zoom_factor*gradient_image.height)))
        return gradient_image_resized

    def CoronalView(self, zoom_factor, iter_gradients, input_size, input_image):
        
        slice_extractor = sitk.ExtractImageFilter()  
        slice_extractor.SetSize([input_size[0], 0, input_size[2]])
        slice_extractor.SetIndex([0, input_size[1]//2, 0])
        extracted_slice = slice_extractor.Execute(input_image)

        gradient_extractor = sitk.VectorIndexSelectionCastImageFilter()
        gradient_extractor.SetIndex(iter_gradients)
        gradient = gradient_extractor.Execute(extracted_slice)
        
        gradient_array = sitk.GetArrayFromImage(gradient)
        gradient_array_normalized = (gradient_array - numpy.min(gradient_array)) * round(255 / numpy.max(gradient_array), 3)
        gradient_array_normalized = numpy.flipud(gradient_array_normalized)
        gradient_image = Image.fromarray(gradient_array_normalized)
        gradient_image = gradient_image.convert("L")
        gradient_image = gradient_image.rotate(90)
        gradient_image_resized = gradient_image.resize((int(zoom_factor*gradient_image.width), int(zoom_factor*gradient_image.height)))
        return gradient_image_resized


from PyQt5.QtWidgets import * 
from PyQt5 import QtCore, QtGui
from PyQt5.QtGui import * 
from PyQt5.QtCore import * 

from dtiplayground.ui.modules_details_communicate import ModulesDetailsCommunicate

class SliceCheck(QWidget):
  communicate = ModulesDetailsCommunicate()

  def __init__(self, protocol_template, slicecheck_yml):
    QWidget.__init__(self)
    self.stack = QWidget()
    self.SliceCheckStack(protocol_template, slicecheck_yml)

  def SliceCheckStack(self, protocol_template, slicecheck_yml):

    ## Module
    self.tab_name = QLabel()

    ## Description
    for template_ite in range(len(protocol_template["options"]["execution"]["pipeline"]["candidates"])):
      if protocol_template["options"]["execution"]["pipeline"]["candidates"][template_ite]["value"] == "SLICE_Check":
        description_label = QLabel(protocol_template["options"]["execution"]["pipeline"]["candidates"][template_ite]["description"])
    description_label.setWordWrap(True)

    ## Options:    
    options_groupbox = QGroupBox("Options:")
    options_layout = QGridLayout()
    options_groupbox.setLayout(options_layout)    
    # overwrite
    self.overwrite = QCheckBox("Overwrite")    
    self.overwrite.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.overwrite, 0, 0)
    # skip
    self.skip = QCheckBox("Skip")
    self.skip.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.skip, 0, 1)
    # write image
    self.writeimage = QCheckBox("Write Image")
    self.writeimage.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.writeimage, 0, 2)
    
    ## Protocol:
    protocol_groupbox = QGroupBox("Protocol:")
    protocol_layout = QFormLayout()
    protocol_groupbox.setLayout(protocol_layout)
    # bSubregionalCheck
    self.bSubregionalCheck_true = QRadioButton("True")
    self.bSubregionalCheck_false = QRadioButton("False")
    bSubregionalCheck_layout = QHBoxLayout()
    bSubregionalCheck_layout.addWidget(self.bSubregionalCheck_true)
    bSubregionalCheck_layout.addWidget(self.bSubregionalCheck_false)
    self.bSubregionalCheck_buttongroup = QButtonGroup()
    self.bSubregionalCheck_buttongroup.addButton(self.bSubregionalCheck_true)
    self.bSubregionalCheck_buttongroup.addButton(self.bSubregionalCheck_false)
    self.bSubregionalCheck_true.toggled.connect(self.GetParams)
    self.bSubregionalCheck_true.setStatusTip(slicecheck_yml["protocol"]["bSubregionalCheck"]["description"])
    self.bSubregionalCheck_false.setStatusTip(slicecheck_yml["protocol"]["bSubregionalCheck"]["description"])
    protocol_layout.addRow(slicecheck_yml["protocol"]["bSubregionalCheck"]["caption"], bSubregionalCheck_layout)
    # subregionalCheckRelaxationFactor:
    self.subregionalCheckRelaxationFactor = QDoubleSpinBox()
    self.subregionalCheckRelaxationFactor.setAlignment(Qt.AlignRight)
    self.subregionalCheckRelaxationFactor.setSingleStep(0.1)
    self.subregionalCheckRelaxationFactor.setDecimals(3)
    self.subregionalCheckRelaxationFactor.setStatusTip(slicecheck_yml["protocol"]["subregionalCheckRelaxationFactor"]["description"])
    protocol_layout.addRow(slicecheck_yml["protocol"]["subregionalCheckRelaxationFactor"]["caption"], self.subregionalCheckRelaxationFactor)    
    self.subregionalCheckRelaxationFactor.valueChanged.connect(self.GetParams)
    # chekTimes
    self.checkTimes = QSpinBox()
    self.checkTimes.setAlignment(Qt.AlignRight)
    self.checkTimes.setStatusTip(slicecheck_yml["protocol"]["checkTimes"]["description"])
    protocol_layout.addRow(slicecheck_yml["protocol"]["checkTimes"]["caption"], self.checkTimes)
    self.checkTimes.valueChanged.connect(self.GetParams)
    # headSkipSlicePercentage
    self.headSkipSlicePercentage = QDoubleSpinBox()
    self.headSkipSlicePercentage.setAlignment(Qt.AlignRight)
    self.headSkipSlicePercentage.setSingleStep(0.1)
    self.headSkipSlicePercentage.setDecimals(3)
    self.headSkipSlicePercentage.setStatusTip(slicecheck_yml["protocol"]["headSkipSlicePercentage"]["description"])
    protocol_layout.addRow(slicecheck_yml["protocol"]["headSkipSlicePercentage"]["caption"], self.headSkipSlicePercentage)   
    self.headSkipSlicePercentage.valueChanged.connect(self.GetParams)
    # tailSkipSlicePercentage
    self.tailSkipSlicePercentage = QDoubleSpinBox()
    self.tailSkipSlicePercentage.setAlignment(Qt.AlignRight)
    self.tailSkipSlicePercentage.setSingleStep(0.1)
    self.tailSkipSlicePercentage.setDecimals(3)
    self.tailSkipSlicePercentage.setStatusTip(slicecheck_yml["protocol"]["tailSkipSlicePercentage"]["description"])
    protocol_layout.addRow(slicecheck_yml["protocol"]["tailSkipSlicePercentage"]["caption"], self.tailSkipSlicePercentage)    
    self.tailSkipSlicePercentage.valueChanged.connect(self.GetParams)
    # correlationDeviationThresholdbaseline
    self.correlationDeviationThresholdbaseline = QDoubleSpinBox()
    self.correlationDeviationThresholdbaseline.setAlignment(Qt.AlignRight)
    self.correlationDeviationThresholdbaseline.setSingleStep(0.1)
    self.correlationDeviationThresholdbaseline.setDecimals(3)
    self.correlationDeviationThresholdbaseline.setStatusTip(slicecheck_yml["protocol"]["correlationDeviationThresholdbaseline"]["description"])
    protocol_layout.addRow(slicecheck_yml["protocol"]["correlationDeviationThresholdbaseline"]["caption"], self.correlationDeviationThresholdbaseline)
    self.correlationDeviationThresholdbaseline.valueChanged.connect(self.GetParams)
    # correlationDeviationThresholdgradient
    self.correlationDeviationThresholdgradient = QDoubleSpinBox()
    self.correlationDeviationThresholdgradient.setAlignment(Qt.AlignRight)
    self.correlationDeviationThresholdgradient.setSingleStep(0.1)
    self.correlationDeviationThresholdgradient.setDecimals(3)
    self.correlationDeviationThresholdgradient.setStatusTip(slicecheck_yml["protocol"]["correlationDeviationThresholdgradient"]["description"])
    protocol_layout.addRow(slicecheck_yml["protocol"]["correlationDeviationThresholdgradient"]["caption"], self.correlationDeviationThresholdgradient)
    self.correlationDeviationThresholdgradient.valueChanged.connect(self.GetParams)
    # quadFit
    self.quadFit_true = QRadioButton("True")
    self.quadFit_false = QRadioButton("False")
    quadFit_layout = QHBoxLayout()
    quadFit_layout.addWidget(self.quadFit_true)
    quadFit_layout.addWidget(self.quadFit_false)
    self.quadFit_buttongroup = QButtonGroup()
    self.quadFit_buttongroup.addButton(self.quadFit_true)
    self.quadFit_buttongroup.addButton(self.quadFit_false)
    self.quadFit_true.toggled.connect(self.GetParams)
    self.quadFit_true.setStatusTip(slicecheck_yml["protocol"]["quadFit"]["description"])
    self.quadFit_false.setStatusTip(slicecheck_yml["protocol"]["quadFit"]["description"])
    protocol_layout.addRow(slicecheck_yml["protocol"]["quadFit"]["caption"], quadFit_layout)    
    
    ## Layout
    layout_v = QVBoxLayout()
    layout_v.addWidget(self.tab_name)
    layout_v.addWidget(description_label)
    layout_v.addWidget(options_groupbox)
    layout_v.addWidget(protocol_groupbox)
    layout_v.addStretch(1)
    self.stack.setLayout(layout_v)


  def GetParams(self):
    params = [
      'Slicewise Check',
      'SLICE_Check', 
      {'options': {
        'overwrite': self.overwrite.isChecked(), 
        'skip': self.skip.isChecked(), 
        'write_image': self.writeimage.isChecked()
        }, 
      'protocol': {
        'bSubregionalCheck': self.bSubregionalCheck_true.isChecked(), 
        'checkTimes': self.checkTimes.value(), 
        'correlationDeviationThresholdbaseline': self.correlationDeviationThresholdbaseline.value(), 
        'correlationDeviationThresholdgradient': self.correlationDeviationThresholdgradient.value(), 
        'headSkipSlicePercentage': self.headSkipSlicePercentage.value(), 
        'quadFit': self.quadFit_true.isChecked(), 
        'subregionalCheckRelaxationFactor': self.subregionalCheckRelaxationFactor.value(), 
        'tailSkipSlicePercentage': self.tailSkipSlicePercentage.value()
        }
      }
    ]
    self.communicate.SendParams(params)



import yaml
from pathlib import Path
from dmri.common.study import loaders 

class Study(object):
  def __init__(self, studies: list, **kwargs):
    self.studies=[] ## study object that contains studies, subjects, sessions, modaliditys and files for each modality

  def load(self, studyfilepath):
    if not Path(studyfilepath).exists(): raise Exception("There is no such file : {}".format(studyfilepath))
    with open(studyfilepath,'r') as f:
      studies=yaml.safe_load(f)
      assert(self.study_validate(studies), "Study file format has something wrong")
      assert(self.check_files(studies), "Some files don't exist")
      self.studies=studies 

  def study_validate(self, study: list):
    is_valid = True
    for s in study:
      for sub in s['subjects']:
        for ses in sub['sessions']:
          for modality in ses:
            print(modality)
    return is_valid


  def from_dir(self, dirpath, storage_format):
    pass

  def generate_dir(self, out_dirpath, storage_format): ### copy files and generate study structure into storage format
    pass

  def check_files(self, study: list):
    for s in study:
      for sub in s['subjects']:
        for ses in sub['sessions']:
          for modality in ses:
            for fn in modality['files']:
              if not Path(fn).exists():
                modality['files_checked']=False
              else:
                modality['files_checked']=True 
    return study



#
#   common/tools/base.py 
#   2021-05-10
#   Written by SK Park, NIRAL, UNC
#
#   External tool wrapper base class
#

import sys
import time
from pathlib import Path 
import subprocess as sp
import dtiplayground.dmri.common 

logger=dtiplayground.dmri.common.logger.write

### decorators

def measure_time(func):  ## decorator
    def wrapper(*args,**kwargs):
        logger("[{}] begins ... ".format(func.__qualname__),dtiplayground.dmri.common.Color.DEV)
        bt=time.time()
        logger("{}".format(args))
        res=func(*args,**kwargs)
        et=time.time()-bt
        logger("[{}] Processed time : {:.2f}s".format(func.__qualname__,et),dtiplayground.dmri.common.Color.DEV)
        return res 
    return wrapper 

class ExternalToolWrapper(object):
    def __init__(self,binary_path):
        self.binary_path=binary_path
        self.arguments=[]
        self.dev_mode=False

    def setDevMode(self,tf:bool):
        self.dev_mode=tf 

    def setPath(self,binary_path : str):
        assert(Path(binary_path).exists())
        self.binary_path=str(binary_path)

    def getPath(self):
        assert(self.binary_path is not None)
        return self.binary_path

    def setArguments(self,args_list:list):
        self.arguments=args_list 

    def getArguments(self):
        return self.arguments 

    def setArgumentsString(self, args_str:str):
        self.arguments=args_str.split()

    def getCommand(self):
        return [self.getPath()]+self.arguments 

    def executeWithArgumentString(self,arguments:str):
        args_list=arguments.split()
        command=[self.getPath()]+self.arguments 
        output=sp.run(command,capture_output=True,text=True)
        #output.check_returncode()
        return output ## output.returncode, output.stdout output.stderr, output.args, output.check_returncode()


    @measure_time
    def execute(self,arguments=None,stdin=None):
        command=self.getCommand()
        if arguments is not None: command=[self.binary_path]+arguments
        output=sp.run(command,capture_output=True,text=True,stdin=stdin)
        if self.dev_mode:
            logger("{}\n{} {}".format(output.args,output.stdout,output.stderr))
            output.check_returncode()
        return output  ## output.returncode, output.stdout output.stderr, output.args, output.check_returncode()
    
    @measure_time
    def execute_pipe(self,arguments=None,stdin=None):
        command=self.getCommand()
        if arguments is not None: command=[self.binary_path]+arguments
        if stdin is None:
            pipe_output=sp.Popen(command,stdout=sp.PIPE)
        else:
            pipe_output=sp.Popen(command,stdin=stdin,stdout=sp.PIPE)
        return pipe_output 


  
import dtiplayground.dmri.preprocessing as prep

logger=prep.logger.write

class MANUAL_Exclude(prep.modules.DTIPrepModule):
    def __init__(self,config_dir,*args,**kwargs):
        super().__init__(config_dir,*args,**kwargs)
        
    def generateDefaultProtocol(self,image_obj):
        super().generateDefaultProtocol(image_obj)
        ## todos
        return self.protocol

    def process(self,*args,**kwargs): ## variables : self.config_dir, self.source_image, self.image (output) , self.result_history , self.result (output) , self.protocol, self.template
        super().process()
        inputParams=self.getPreviousResult()['output']
        # << TODOS>>

        self.result['output']['excluded_gradients_original_indexes']= self.protocol['gradientsToExclude'] ## this is code to exclude the gradients 
        self.result['output']['success']=True
        return self.result



from dtiplayground.dmri.common.tools.base import ExternalToolWrapper

class DTIEstim(ExternalToolWrapper):
    def __init__(self,binary_path):
        super().__init__(binary_path)
    
    def estimate(self,dwi_path, output_path, options=[]):
        arguments=[]
        arguments+=['--dwi_image',dwi_path]
        arguments+=['--tensor_output',output_path]
        arguments+=options
        self.setArguments(arguments)
        return self.execute(arguments)
        




  
import dtiplayground.dmri.preprocessing as prep
from dtiplayground.dmri.preprocessing.dwi import DWI
import dtiplayground.dmri.common
import yaml
from pathlib import Path 
import SUSCEPTIBILITY_Correct.utils as utils
from dtiplayground.dmri.common import measure_time
import dtiplayground.dmri.common.tools as tools 
import shutil
import copy
logger=prep.logger.write


class SUSCEPTIBILITY_Correct(prep.modules.DTIPrepModule):
    def __init__(self,config_dir,*args,**kwargs):
        super().__init__(config_dir,*args,**kwargs)


    def install(self,install_dir,*args,**kwargs):
        topup_config='fsl_regb02b0.cnf'
        install_dir=Path(install_dir).joinpath('parameters').joinpath('fsl')
        install_dir.mkdir(parents=True,exist_ok=True)
        source_path=Path(dtiplayground.dmri.common.__file__).parent.joinpath('data').joinpath(topup_config)
        dest_path=install_dir.joinpath(topup_config)
        logger("Copying FSL config file to {}".format(str(dest_path)))
        shutil.copy(source_path,dest_path)

    def generateDefaultEnvironment(self):
        return super().generateDefaultEnvironment()
    
    def checkDependency(self,environment): #use information in template, check if this module can be processed
        # FSL should be ready before execution
        software_path=Path(self.config_dir).joinpath('software_paths.yml')
        software_info = yaml.safe_load(open(software_path,'r'))
        if self.name in environment:
            fslpath=Path(software_info['softwares']['FSL']['path'])
            try:
                #fslpath=Path(environment[self.name]['fsl_path'])
                fsl_exists=fslpath.exists()
                if fsl_exists:
                    return True, None 
                else:
                    return False, "FSL Path doesn't exist : {}".format(str(fslpath))
            except Exception as e:
                return False, "Exception in finding FSL6 : {}".format(str(e))
        else:
            return False, "Can't locate FSL" #test

    def generateDefaultProtocol(self,image_obj):
        super().generateDefaultProtocol(image_obj)
        ## todos
        self.protocol['configurationFilePath']=self.protocol['configurationFilePath'].replace("$CONFIG_DIR",str(self.config_dir))
        return self.protocol

    def process(self,*args,**kwargs): ## variables : self.global_variables, self.softwares, self.output_dir, self.config_dir, self.source_image, self.image (output) , self.result_history , self.result (output) , self.protocol, self.template
        super().process()
        #inputParams=self.getPreviousResult()['output']
        gradient_indexes_to_remove=[]
        protocol_options=args[0]
        self.num_threads=protocol_options['software_info']['parameters']['num_max_threads']
        self.software_info=protocol_options['software_info']['softwares']
        self.baseline_threshold=protocol_options['baseline_threshold']

        res=None
        output_image_file=Path(self.output_dir).joinpath('output.nrrd')
        logger("Running topup ...",prep.Color.PROCESS)
        self.run_topup( phaseEncodingAxis=self.protocol['phaseEncodingAxis'],
                        phaseEncodingValue=self.protocol['phaseEncodingValue'],
                        configurationFilePath=self.protocol['configurationFilePath'])
        logger("Topup done",prep.Color.OK)

        ## results
        self.result['output']['excluded_gradients_original_indexes']=self.image.convertToOriginalGradientIndex(gradient_indexes_to_remove)
        self.result['output']['success']=True
        return self.result

### User defined methods
### scripts
    @measure_time
    def convert_to_nifti(self,phaseEncodingAxis): ## generate nifti file and return filenames
        axis_alias={0:'lr',1:'ap',2:'is'}
        ped=axis_alias[phaseEncodingAxis[0]]
        inv_ped=ped[::-1] #reverse string
        directions=[ped,inv_ped]
        out_files=[]
        out_images=[]
        if ped=='ap' : directions+=['fh','hf']
        for d in directions:
            for img in self.images:
                fname=Path(img.filename).name.lower()
                if d in fname:
                    outfilename=Path(self.output_dir).joinpath(d+".nii.gz").__str__()
                    img.image_type='nifti'
                    img.writeImage(outfilename,dest_type='nifti')
                    out_files.append(outfilename)
                    out_images.append(img)
        if len(out_files)<2:
            out_files=[]
            for idx,img in enumerate(self.images):
                outfilename=Path(self.output_dir).joinpath(directions[idx]+".nii.gz").__str__()
                img.writeImage(outfilename,dest_type='nifti')
                out_files.append(outfilename)
                out_images.append(img)

        return out_files,out_images


    @measure_time
    def create_b0s(self,pe_images,b0_threshold):  ## create b0 and index file
        b0_files=[]
        b0_images=[]
        b0_indexes=[]
        for img in pe_images:
            minb,maxb=img.getBValueBounds()
            if b0_threshold < minb :   ## if given b0 threshold is less than the least b value of the gradients, substitute it to the minimal b value
                img.setB0Threshold(minb)
                b0_threshold=minb

        for idx,img in enumerate(pe_images):

            b0img=img.extractBaselines(b0_threshold)
            outfilename=Path(self.output_dir).joinpath("b0_{}.nii.gz".format(idx)).__str__()
            index_outfilename=Path(self.output_dir).joinpath("b0_{}.index".format(idx)).__str__()
            b0img.writeImage(outfilename,dest_type='nifti')
            grads=b0img.getGradients()
            indexes=[str(x['original_index']) for x in grads]
            with open(index_outfilename,'w') as f :
                f.write(" ".join(indexes))
            b0_files.append(outfilename)
            b0_images.append(b0img)
            b0_indexes.append(index_outfilename)
        return b0_files,b0_images,b0_indexes,b0_threshold

    def merge_images(self,outputfilename,pe_files:list,b0_threshold):
        fsl=tools.FSL(self.software_info['FSL']['path'])
        fsl._set_num_threads(self.num_threads)
        
        output=fsl.fslmerge(outputfilename,pe_files)
        ## making merged bvals,bvecs
        bvals_fn=Path(self.output_dir).joinpath(Path(outputfilename).name.split('.')[0]+'.bval')
        bvecs_fn=Path(self.output_dir).joinpath(Path(outputfilename).name.split('.')[0]+'.bvec')

        input_bvals_fn_0=Path(self.output_dir).joinpath(Path(pe_files[0]).name.split('.')[0]+'.bval')
        input_bvecs_fn_0=Path(self.output_dir).joinpath(Path(pe_files[0]).name.split('.')[0]+'.bvec')
        input_bvals_fn_1=Path(self.output_dir).joinpath(Path(pe_files[1]).name.split('.')[0]+'.bval')
        input_bvecs_fn_1=Path(self.output_dir).joinpath(Path(pe_files[1]).name.split('.')[0]+'.bvec')
        
        with open(bvals_fn,'w') as fw:
            bvals1=open(input_bvals_fn_0,'r').read()
            bvals2=open(input_bvals_fn_1,'r').read()
            bvals=bvals1+bvals2
            bvals_int=list(map(int,bvals.split()))
            for idx,b in enumerate(bvals_int):
                if b<=b0_threshold:
                    bvals_int[idx]=0 ## fsl's baseline should have 0 b values
            bvals="\n".join(list(map(str,bvals_int)))
            fw.write(bvals)

        with open(bvecs_fn,'w') as fw:
            bvecs1=open(input_bvecs_fn_0,'r').read()
            bvecs2=open(input_bvecs_fn_1,'r').read()
            bvecs=bvecs1+bvecs2
            fw.write(bvecs)


        logger(output.stdout)
        logger(output.stderr)
        output.check_returncode()


    ### fsl parameters
    def make_acqp(self,outfilename,b0index_files:list,axis=1,val=0.0924):
        phase_dir=[0,0,0,val]
        result=[]
        direction=1
        for index_file in b0index_files:
            indexes=open(index_file,'r').read().split()
            for i in indexes:
                p=copy.deepcopy(phase_dir)
                p[axis]=direction
                result.append(p)
            direction*=-1
        with open(outfilename,'w') as fw:
            for r in result:
                strline="{:d} {:d} {:d} {:.4f}\n".format(*r)
                fw.write(strline)
        return result

    def make_index(self,image,outfilename):
        indices=image.getB0Index()
        with open(outfilename,'w') as fw:
            outstr=" ".join(list(map(str,indices)))
            fw.write(outstr)
        return indices 

    def get_phase_axis(self,phaseEncodingAxis):
        return phaseEncodingAxis[0]

    def zero_padding_odd_sizes(self):
        for img in self.images:
            pad_sizes=[0,0,0,0]
            for idx,x in enumerate(list(img.images.shape)[0:3]):
                if x%2==1: 
                    logger("[WARNING] Image size has odd number, automatically padding zero values...",prep.Color.WARNING)
                    pad_sizes[idx]+=1 
            img.zeroPad(pad_sizes)

    @measure_time
    def run_topup(  self,
                    phaseEncodingAxis,
                    phaseEncodingValue,
                    configurationFilePath):

        b0_threshold=self.baseline_threshold 
        phase_axis=self.get_phase_axis(phaseEncodingAxis)

        output_dir=Path(self.output_dir)
        output_nrrd=output_dir.joinpath('output.nrrd').__str__()
        
        fsl=tools.FSL(self.software_info['FSL']['path'])
        fsl._set_num_threads(self.num_threads)
        fsl.setDevMode(True) 

        ## auto zero padding if image dimension has odd number of element (even size of image sizes is only acceptable)
        logger("Checking image sizes...",prep.Color.INFO)
        self.zero_padding_odd_sizes()


        pe_files, pe_images=self.convert_to_nifti(phaseEncodingAxis)
        b0_files, b0_images, b0_index_files,b0_threshold=self.create_b0s(pe_images,b0_threshold)
        
        logger("Merging b0s ...",prep.Color.PROCESS)
        merged_b0_filename=Path(self.output_dir).joinpath("b0_merged.nii.gz").__str__()
        self.merge_images(merged_b0_filename,b0_files,b0_threshold)

        logger("Generating acqp parameters ...",prep.Color.PROCESS)
        acqp_filename=Path(self.output_dir).joinpath("acqp.txt").__str__()
        self.make_acqp(acqp_filename, b0_index_files,axis=phase_axis,val=phaseEncodingValue)

        logger("Merging original images ... ",prep.Color.PROCESS)
        merged_image_filename=Path(self.output_dir).joinpath("merged_image.nii.gz").__str__()
        merged_bvals_filename=Path(self.output_dir).joinpath("merged_image.bval").__str__()
        merged_bvecs_filename=Path(self.output_dir).joinpath("merged_image.bvec").__str__()
        self.merge_images(merged_image_filename,pe_files,b0_threshold)

        ## load merged image into self.image 
        self.image=DWI(merged_image_filename)
        self.image.setB0Threshold(b0_threshold)
        base=Path(merged_image_filename).name.split('.')[0]

        logger("Generating index file ...",prep.Color.PROCESS)
        _index_path=Path(self.output_dir).joinpath(base+".b0index").__str__()
        indices=self.make_index(self.image,_index_path)
        logger("B0 indices : {}".format(indices),prep.Color.INFO)

        logger("Running topup command...",prep.Color.PROCESS)
        _out_path=Path(self.output_dir).joinpath(base+"_topup_fieldcoef.nii.gz").__str__()
        _fout_path=Path(self.output_dir).joinpath(base+"_field.nii.gz").__str__()
        _iout_path=Path(self.output_dir).joinpath(base+"_corr.nii.gz").__str__()
        _out_path_prep=Path(self.output_dir).joinpath(base+"_topup").__str__()
        if not Path(_out_path).exists() or not Path(_fout_path).exists() or not Path(_iout_path).exists():
            
            output=fsl.topup(imain=merged_b0_filename,  # input image filename (merged b0)
                      datain=acqp_filename, # acqp params filename
                      out=_out_path_prep,    # output basename (not a filename)
                      fout=_fout_path,   # field output filename (Hz)
                      iout=_iout_path,   # movement corrected image output filename
                      config=configurationFilePath) # config filename
        
        _average_path=Path(self.output_dir).joinpath(base+"_average.nii.gz").__str__()
        
        logger("Averaging topup corr for the masking...",prep.Color.PROCESS)
        output=fsl.fslmaths_ops(_iout_path,_average_path,'mean')
        logger("Generating mask...",prep.Color.PROCESS)
        _mask_path=Path(self.output_dir).joinpath(base+"_mask.nii.gz").__str__()
        output=fsl.bet(_average_path,_mask_path)
        _mask_path=Path(self.output_dir).joinpath(base+"_mask_mask.nii.gz").__str__()
        
        self.image.zerorizeBaselines(b0_threshold) ## fsl's baseline should have 0 b values

        susceptibility_parameters={
            "mask_path" : _mask_path,
            "acqp_path" : acqp_filename,
            "image_path" : merged_image_filename,
            "image_bvals_path" : merged_bvals_filename,
            "image_bvecs_path" : merged_bvecs_filename,
            "b0_image_path" : merged_b0_filename,
            "topup_path" : _out_path_prep,
            "index_path" : _index_path
        }
        self.updateGlobalVariable(susceptibility_parameters)

        self.result['output']['susceptibility_parameters']=susceptibility_parameters
        self.writeImageWithOriginalSpace(output_nrrd,'nrrd')
